{ PktXCode - decodes with UUENCODE, XXENCODE, BASE64 or BINHEX encoded files
             out of text files, Fido-PKT-/MSG- or ZConnect files

  Copyright (C) 1996-98 Ulrich Mueller, email: ulrich.mueller@bigfoot.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


  This source code is compatible with:
  - Borland Pascal 7.0 (DOS)
  - Virtual Pascal 1.1 (OS/2)

}

Unit BinkOut;

{$I SETTINGS.INI}

{ Unit zur Benutzung des Binkley-Outbounds, Revision Alpha 0003
  Copyright (C) 1996-98 Tobias Ernst, Fido: Tobias Ernst@2:2476/418

  Dokumentation siehe Binkout.Txt
  modifiziert fr Schnittstellen von PKTX_LFN/PKTX_SCR von Ulrich Mueller
  Diese Unit untersteht der GNU General Public License.
}

INTERFACE

Type T_Nodedata = Record
                     zone:   integer;
                     net:    integer;
                     node:   integer;
                     point:  integer;
                     is5d:   boolean;
                     domain: String;
                   end;

const ST_NORMAL    = 0;
const ST_KILLSENT  = 64;
const ST_TRUNCSENT = 65;

const FL_NORMAL    = 0;
const FL_CRASH     = 1;
const FL_HOLD      = 2;
const FL_DIRECT    = 3;
const FL_IMMEDIATE = 4;

const BOUT_NOERROR      = 0;
      BOUT_NODEISBUSY   = 1;
      BOUT_WRITEERROR   = 2;
      BOUT_READERROR    = 3;
      BOUT_UNCONFIGURED = 4;
      BOUT_OUTOFMEMORY  = 5;

var BOUT_Result: integer;

Procedure BOUT_ConfigureDomain (nZone: integer; strDomainname: String; strOutboundname: String);
Procedure BOUT_ConfigureBasepath (strBasepathname: String);
Procedure BOUT_ConfigureHomeZone (zone: integer);
Procedure BOUT_InsertFile (strFilespec: String; NodeData: T_Nodedata; flags: integer);
Procedure BOUT_RemoveFile (strfilespec: String; NodeData: T_Nodedata);

IMPLEMENTATION

uses pktx_lfn, pktx_scr, dos; { fr FSplit }

{ Die Domain-Information (Zuordnung Zonenummer<->Domain) wird in einer
  einfach verketteten Liste festgehalten.
}

Type prDomain = ^rDomain;
     rDomain  = Record
                  nZone: integer;
                  strDomainname: String;
                  strOutboundname: String;
                  next: prDomain;
                end;

const FL_TMP       = 13;
{ Name einer temporren Datei, nur fr internen Gebrauch }
const FL_BUSY      = 14;
{ Name der Busy-Semaphore, nur fr internen Gebrauch }
const FL_ANY       = 15;
{ ? als Flavourbuchstabe, nur fr internen Gebrauch }

var strBasepath: String;
    strDefaultOutbound: String;
    pDomainRoot: prDomain;
    nHomeZone: integer;



Function IsConfigured: Boolean;
begin
  IsConfigured:=((nHomeZone>0)AND(Length(strBasepath)>0));
end;


Procedure BOUT_ConfigureBasepath (strBasepathname: String);
  { Den vom Benutzer angegebenen Default-Outbound (z.B. e:\mailer\bink\out)
    zerlegen in Outboundbasispfad (e:\mailer\bink\) und Default-Outboundname
    (out)
  }
begin
  BOUT_Result:=BOUT_NOERROR;

  strBasepathname   :=fnShortPath(strBasepathname);
  strBasePath       :=fnDir(strBasepathname);
  strDefaultOutbound:=fnName(strBasepathname);
end;


Procedure BOUT_ConfigureDomain (nZone: integer; strDomainname: String; strOutboundname: String);
var pNew, p: prDomain;
begin

  BOUT_Result:=BOUT_NOERROR;

  new(pNew); if (pNew=NIL) then begin
                             BOUT_Result:=BOUT_OUTOFMEMORY;
                             exit;
                            end;

  pNew^.nZone:=nZone;
  pNew^.strDomainname:=strDomainname;
  pNew^.strOutboundname:=strOutboundname;
  pNew^.next:=NIL;

  if (pDomainRoot=NIL)
    then pDomainRoot:=pNew
    else begin
      p:=pDomainRoot;
      while (p^.next<>NIL) do p:=p^.next;
      p^.next:=pNew;
    end;

end;



Procedure BOUT_ConfigureHomeZone (zone: integer);
begin
  BOUT_Result:=BOUT_NOERROR;
  nHomeZone:=zone;
end;

{ Bastelt den Namen des ?LO-Files zusammen. Eventuell nicht existierende
  Verzeichnisse werden gleich mit angelegt.
}
function LO_Filename (NodeData: T_Nodedata; flags: integer): String;

  { Diese Routine gibt Bldsinn zurck, wenn n>=(16 hoch width) }
  function printhex(n: longint; width: longint): String;
  const hexdigits: array[0..15] of char =  '0123456789ABCDEF';
  var str: String;
      ndigits: longint;
  begin
    str:='';
    for ndigits:=0 to width-1 do begin
      str:=hexdigits[n mod 16]+str;
      n:=n div 16;
    end;
    printhex:=str;
  end;

var strLO: String;
    strOutbound: String;
    p: prDomain;
    DirInfo: SearchRec;

begin

  strLO:=strBasePath;                 { z.B. e:\bink\ }
  if (strLO<>'') then begin

    strOutbound:=strDefaultOutbound;
    p:=pDomainRoot;
    while (p<>NIL) do begin
      if (      (NodeData.zone=p^.nZone) AND
                (      (upcasestr(NodeData.domain)=upcasestr(p^.strDomainname))
                    OR (NOT Nodedata.is5d)
                )
         )
        then begin
          strOutbound:=p^.strOutboundname;
          p:=NIL;
        end else p:=p^.next;
    end;

    strLO:=strLO+strOutbound;     { z.B. e:\bink\out }
    if (NodeData.zone<>nHomeZone)
      then strLO:=strLo+'.'+printHex(NodeData.zone,3);
                                  { z.B. e:\bink\out oder
                                         e:\bink\out.0f2 }

    if not lfnexistpath(strlo) then lfncreatepath(strlo);

    strLO:=strLO+'\'+printHex(NodeData.net,4)+printHex(NodeData.node,4)+'.';
                                  { z.B. e:\bink\out\09ac01a2. };

    if (NodeData.point<>0) then begin
      strLO:=strLO+'PNT';        { z.B. e:\bink\out\09ac01a2.pnt }

      if not lfnexistpath(strlo) then lfncreatepath(strlo);

      strLO:=strLO+'\'+printHex(NodeData.point,8)+'.'
                                 { z.B. e:\bink\out\09ac01a2.pnt\0000000E. }
    end;

    case (flags AND 15) of
         FL_CRASH:     strLO:=strLO+'CLO';
         FL_HOLD:      strLO:=strLO+'HLO';
         FL_DIRECT:    strLO:=strLO+'DLO';
         FL_IMMEDIATE: strLO:=strLO+'ILO';
         FL_NORMAL:    strLO:=strLO+'FLO';
         FL_TMP:       strLO:=strLO+'TMP';
         FL_BUSY:      strLO:=strLO+'BSY';
         FL_ANY:       strLO:=strLO+'?LO';
         else          strLO:=strLO+'FLO';  { oder Fehler zurckgeben ... }
    end;
                                 { z.B. e:\bink\out\09ac01a2.hlo }
  end;
  LO_Filename:=strLO;
end;



Function BOUT_Lock(NodeData: T_Nodedata): Boolean;
var f: file;
    r: boolean;
    fname : string;
begin
  fname:=LO_Filename(NodeData,FL_BUSY);

  if not lfnexistfile(fname) then begin
    lfnopenfile(f, fname, ofcreate);
    lfnclosefile(f);
    r:=TRUE;
  end else r:=FALSE;

  BOUT_Lock:=r;

end;



Procedure BOUT_Unlock(NodeData: T_Nodedata);
begin
  lfnerasefile(LO_Filename(NodeData,FL_BUSY));
end;



Procedure BOUT_InsertFile (strFilespec: String; NodeData: T_Nodedata; flags: integer);
var strFlowfile: String;
    hFlowfile: text;
begin

  if (NOT IsConfigured) then begin
    BOUT_Result:=BOUT_UNCONFIGURED;
    exit;
  end;

  if (NOT BOUT_Lock(NodeData)) then begin
    BOUT_Result:=BOUT_NODEISBUSY;
    exit;
  end;

  BOUT_Result:=BOUT_NOERROR;

  strFlowfile:=LO_Filename(NodeData,flags);
  assign(hFlowfile,strFlowfile);
  {.$I-}
  append(hFlowfile);
  {.$I+}
  geterror;

  if (Result<>0) then begin
    {.$I-}
    rewrite(hFlowfile);
    geterror;
    {.$I+}
    if (Result<>0) then begin
      BOUT_Result:=BOUT_WRITEERROR;
      BOUT_Unlock(NodeData);
      exit;
    end;
  end;

  case (flags AND (NOT 63)) of
    ST_KILLSENT:  strFilespec:='^'+strFilespec;
    ST_TRUNCSENT: strFilespec:='#'+strFilespec;
    else;
  end;

  {.$I-}
  writeln(hFlowfile,strFilespec);
  geterror;
  {.$I+}
  if (Result<>0) then BOUT_Result:=BOUT_WRITEERROR;
  {.$I-}
  close(hFlowfile);
  geterror;
  {.$I+}
  if (Result<>0) then BOUT_Result:=BOUT_WRITEERROR;

  BOUT_Unlock(NodeData);
end;



Procedure BOUT_RemoveFile (strfilespec: String; NodeData: T_Nodedata);
var strFlowfile,s,s2,strTemp: String;
    hFlowfile,hTemp: text;
    DirInfo: SearchRec;
    nLinesWritten: integer;
    bModified: boolean;
    bTempCreated: boolean;
begin

  if (NOT IsConfigured) then begin
    BOUT_Result:=BOUT_UNCONFIGURED;
    exit;
  end;

  if (NOT BOUT_Lock(NodeData)) then begin
    BOUT_Result:=BOUT_NODEISBUSY;
    exit;
  end;

  BOUT_Result:=BOUT_NOERROR;

  strFlowfile:=LO_Filename(NodeData,FL_ANY);
  bTempCreated:=FALSE;
  strTemp:=LO_Filename(NodeData,FL_TMP);
  FindFirst(strflowfile,Anyfile,DirInfo);
  while DosError = 0 do
  begin
    {.$I-}
    strflowfile[Length(strflowfile)-2]:=DirInfo.Name[Length(DirInfo.Name)-2];
    assign(hFlowfile,strFlowfile);
    assign(hTemp,strTemp);
    reset(hFlowfile);
    geterror;
    if (Result<>0) then begin
      BOUT_Result:=BOUT_READERROR;
      BOUT_Unlock(NodeData);
      exit;
    end;
    rewrite(hTemp);
    geterror;
    if (Result<>0) then begin
      BOUT_Result:=BOUT_WRITEERROR;
      close(hFlowFile);
      BOUT_Unlock(NodeData);
      exit;
    end;
    bModified:=FALSE;
    bTempCreated:=TRUE;
    nLinesWritten:=0;

    while not Eof(hFlowfile) do
    begin
      Readln(hFlowfile,s);
      geterror;
      if (Result<>0) then begin
        BOUT_Result:=BOUT_READERROR;
        close(hFlowFile);
        close(hTemp);
        BOUT_Unlock(NodeData);
        exit;
      end;

      s2:=s;
      if (upcasestr(s2)<>upcasestr(strfilespec)) then
        delete(s2,1,1);
      if (upcasestr(s2)<>upcasestr(strfilespec)) then begin
        inc(nLinesWritten);
        Writeln(hTemp,s);
        geterror;
        if (Result<>0) then begin
          BOUT_Result:=BOUT_WRITEERROR;
          close(hFlowFile);
          close(hTemp);
          BOUT_Unlock(NodeData);
          exit;
        end;
      end else
        bModified:=TRUE;
    end;

    Close(hFlowfile);
    geterror;
    if (Result<>0) then begin
      BOUT_Result:=BOUT_READERROR;
      BOUT_Unlock(NodeData);
      close(hTemp);
      exit;
    end;

    Close(hTemp);
    geterror;
    if (Result<>0) then begin
      BOUT_Result:=BOUT_WRITEERROR;
      BOUT_Unlock(NodeData);
      exit;
    end;

    if (bModified) then begin
      Erase(hFlowfile);
      geterror;
      if (Result<>0) then begin
        BOUT_Result:=BOUT_WRITEERROR;
        BOUT_Unlock(NodeData);
        exit;
      end;
      if (nLinesWritten>0) then begin
        Rename(hTemp,strFlowfile);
        geterror;
      end;
      if (Result<>0) then begin
        BOUT_Result:=BOUT_WRITEERROR;
        BOUT_Unlock(NodeData);
        exit;
      end;
    end;
    { hTemp wird erst am Ende der Schleife gelscht, um Eintrge im
      DELDIR zu minimieren }
    {.$I+}
    FindNext(DirInfo);
  end;

  if (bTempCreated) then begin
    lfnerasefile(strTemp);
    if (Result<>0) then begin
      BOUT_Result:=BOUT_WRITEERROR;
      BOUT_Unlock(NodeData);
      exit;
    end;
  end;

  BOUT_Unlock(NodeData);
end;


begin
  pDomainRoot:=NIL;
  strBasePath:='';
  nHomeZone:=-1;
  BOUT_Result:=BOUT_NOERROR;
end.