{ PktXCode - decodes with UUENCODE, XXENCODE, BASE64 or BINHEX encoded files
             out of text files, Fido-PKT-/MSG- or ZConnect files

  Copyright (C) 1996-98 Ulrich Mueller, email: ulrich.mueller@bigfoot.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


  This source code is compatible with:
  - Borland Pascal 7.0 (DOS)
  - Virtual Pascal 1.1 (OS/2)

}

Unit PKTX_DEC;

{$I SETTINGS.INI}

{ 12.01.97 Erstellt
           Paralleldekodierung
  26.01.97 UUDecode kommt mit "kurzen" Zeilen zurecht
  07.02.97 keine CaseSensity bei BASE64-Header
           Fido-FSC-0047 kompatible Split-Erkennung (Betreff egal!)
  08.02.97 Untersttzung von UU-Split-Sections (Part-Ersatz)
           Base64: letzte zeile kann n*4 + "="-Zeichen haben
  06.03.97 bei UUorXX-Start wurde nField nicht initalisiert
  09.04.97 bei UUorXX-Start wurde CloseOutField nicht aufgerufen
  27.06.97 fehlerhafte Fido-FSC-Implementation wird untersttzt
  26.09.97 Ende eines PGP-Blocks wird korrekt erkannt
  19.12.97 Dateinamen werden auf Pfadangaben berprft
  05.11.98 Fido-Seenby-Zeilen werden vom Dekodieren ausgeschlossen
}

INTERFACE

Uses PKTX_IO,
     PKTX_SCR,
     PKTX_KEY,
     PKTX_DB,
     PKTX_MEM,
     PKTX_LFN
 {$IFDEF Binkley}
   , PKTX_BNK
 {$ENDIF};

{ diese Unit decodiert die einzelnen Textzeilen und nimmt sie in die
  Datenbank auf. Unter bestimmten Umstnden modifiziert sie den OutMailBuffer

  deInit und deClose sollen vor und nach jeder Nachricht aufgerufen werden

  Zugriff auf:
    OutString   - enthlt die zu decodierende Zeile
    OutField^   - wird zur Pufferung der Dekodierung benutzt
    OutPos      - Zeiger innerhalb OutPos
    ThisMsgData - true, wenn Nachricht Daten enthielt
}

{ Verbesserungen:
    alle Dekodierroutinen knnen parallel benutzt werden
    nur einmal Reduzierung der Zeilenlnge zulssig

  Problem:
    einzelne Zeile mit Endmarkierung wird nur im UU/XX-Format erkannt

}

PROCEDURE deInit;
PROCEDURE deDecode;
PROCEDURE deClose;
PROCEDURE deGetMem;
PROCEDURE deFreeMem;

FUNCTION OutStringStart_CaseSensity(Var Buffer): Boolean;

IMPLEMENTATION

{***************************************************************************}

CONST uNormalLineSize = 45*3;
      xNormalLineSize = 45*3;
      bNormalLineSize = 45*3;
      hNormalLineSize = 64*3;
      CodeLineSize    = 40; { alles drunter mu lastline sein ! }
      MinCodeSize     = 30; { alles drunter mu lastline sein ! }

      DestEndUnknown  = $FFFFFFFF;

CONST mtlInit       = 0;
      mtlLeerzeile  = 1;
      mtlData       = 2;

CONST daBufSize   = 4096;
      declen_init = 255;

TYPE tdaBuf      = Array[0..daBufSize-1] Of Char;
TYPE tDecodeArea = Record
                     Buf        : PField;  { Puffer fr dekodete Daten }
                     OrgBuf     : ^tdaBuf; { Puffer fr dekodete Daten }
                     BufPos     : Word;    { Zeiger in Puffer }
                     DecLen     : Byte;    { Lnge der zuletzt kod. Daten }
                     LastLine   : Boolean; { letzte (=krzere) Zeile gefund }
                     DecOpen    : Boolean; { bereits eine Zeile dekodiert }
                     DestStart  : LongInt; { Startadresse in DestFile }
                     DestEnd    : LongInt;
                   End;
     pDecodeArea = ^tDecodeArea;

VAR uField, xField,
    bField, hField      : tDecodeArea;
    nField              : pDecodeArea;  { nField: Field des akt. Typs }

CONST deInitialized : Boolean = False;

VAR PGPBlockDetected    : Boolean;
    MimeDetected        : Boolean;
    MimeTextLevel       : Byte;
    MimeTextStart,
    MimeTextEnd         : LongInt;
    DiffPos             : Word; { dient zur Bestimmung der Dateigre }

    FirstNotDecodedLine_Pos : LongInt; { wird benutzt, wenn OutIDX.Typ<>0
                                         zeigt erste Zeile an, die nicht
                                         dekodierbar war }

{***************************************************************************}
{ Umrechnungstabellen }
{***************************************************************************}

CONST XXField: Array[0..$FF] Of Byte=  {umrechnungstabelle xxdec->bytes}
      { 255: ungltiges Zeichen }
{000}  ($FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{016}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{032}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$00,$FF,$01,$FF,$FF,
{048}   $02,$03,$04,$05,$06,$07,$08,$09,$0A,$0B,$FF,$FF,$FF,$FF,$FF,$FF,
{064}   $FF,$0C,$0D,$0E,$0F,$10,$11,$12,$13,$14,$15,$16,$17,$18,$19,$1A,
{080}   $1B,$1C,$1D,$1E,$1F,$20,$21,$22,$23,$24,$25,$FF,$FF,$FF,$FF,$FF,
{096}   $FF,$26,$27,$28,$29,$2A,$2B,$2C,$2D,$2E,$2F,$30,$31,$32,$33,$34,
{112}   $35,$36,$37,$38,$39,$3A,$3B,$3C,$3D,$3E,$3F,$FF,$FF,$FF,$FF,$FF,
{128}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{144}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{160}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{176}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{192}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{208}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{224}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{240}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF);

      Base64: Array[0..$FF] Of Byte=  {umrechnungstabelle base64->bytes}
      { 255: ungltiges Zeichen, 128: hier zeilenende }
{000}  ($FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{016}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{032}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$3E,$FF,$FF,$FF,$3F,
{048}   $34,$35,$36,$37,$38,$39,$3A,$3B,$3C,$3D,$FF,$FF,$FF,$80,$FF,$FF,
{064}   $FF,$00,$01,$02,$03,$04,$05,$06,$07,$08,$09,$0A,$0B,$0C,$0D,$0E,
{080}   $0F,$10,$11,$12,$13,$14,$15,$16,$17,$18,$19,$FF,$FF,$FF,$FF,$FF,
{096}   $FF,$1A,$1B,$1C,$1D,$1E,$1F,$20,$21,$22,$23,$24,$25,$26,$27,$28,
{112}   $29,$2A,$2B,$2C,$2D,$2E,$2F,$30,$31,$32,$33,$FF,$FF,$FF,$FF,$FF,
{128}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{144}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{160}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{176}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{192}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{208}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{224}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{240}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF);

      BinHexField: Array[0..$FF] Of Byte=  {umrechnungstabelle Binhex->Bytes}
      { 255: ungltiges Zeichen, 128: ":"-Markierung }
{000}  ($FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{016}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{032}   $FF,$00,$01,$02,$03,$04,$05,$06,$07,$08,$09,$0A,$0B,$0C,$FF,$FF,
{048}   $0D,$0E,$0F,$10,$11,$12,$13,$FF,$14,$15,$80,$FF,$FF,$FF,$FF,$FF,
{064}   $16,$17,$18,$19,$1A,$1B,$1C,$1D,$1E,$1F,$20,$21,$22,$23,$24,$FF,
{080}   $25,$26,$27,$28,$29,$2A,$2B,$FF,$2C,$2D,$2E,$2F,$FF,$FF,$FF,$FF,
{096}   $30,$31,$32,$33,$34,$35,$36,$FF,$37,$38,$39,$3A,$3B,$3C,$FF,$FF,
{112}   $3D,$3E,$3F,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{128}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{144}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{160}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{176}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{192}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{208}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{224}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,
{240}   $FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF);

{***************************************************************************}
{ UU-Decodierung }
{***************************************************************************}

FUNCTION UUDecodAble: Boolean; Assembler;
{dekodiert outstring in outfield, wenn mglich}
ASM
{ Zeiger initalisieren }
{$IFDEF OS2} push esi                     {$ELSE} push ds                    {$ENDIF}
{$IFDEF OS2} push edi                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} push ebx                     {$ELSE}                            {$ENDIF}

{$IFDEF OS2} mov edi, ufield.buf          {$ELSE} les di, ufield.buf         {$ENDIF}
{$IFDEF OS2} movzx eax,  ufield.bufpos    {$ELSE} mov ax, ufield.bufpos      {$ENDIF}
{$IFDEF OS2} add edi, eax                 {$ELSE} add di, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov esi, offset outstring    {$ELSE} mov ax, seg outstring      {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov ds, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov si, offset outstring   {$ENDIF}

             cld
             lodsb
{$IFDEF OS2} xor ecx, ecx                 {$ELSE} xor cx, cx                 {$ENDIF}
             mov cl, al
             dec cl      { cl: Anzahl Textzeichen (Quelle) }

             { Gre testen }
             lodsb
             sub al, 32
             jbe @not_decodable
             cmp al, 64
             ja @not_decodable
             jb @0

             cmp cl, 0
             jne @not_decodable

             xor dh, dh
             mov ufield.declen, 0    { sonst Probleme bei firstline }
             jmp @decodable_lastline { Zeile mit "`" }

           @0:
             mov ah, ufield.lastline
             or  ah, ah
             jnz @not_decodable

             mov dh, al  { dh: Anzahl der Zeichen, die entstehen sollen }

             { Gre mit voriger Zeile vergleichen }

             mov ah, ufield.declen
             cmp ah, declen_init
             je @line_ok

             cmp dh, ah
             ja @not_decodable

           @line_ok:
             xor ah, ah  { *4, div 3, +1? }
             shl ax, 2

             mov dl, 3
             div dl       { ax div dl }
             { ah: mod }
             or  ah, ah
             jz @1
             inc al
           @1: { al: Anzahl der dafr bentigten Textzeichen }

             cmp al, cl
             ja @not_decodable { al>cl }
             je @3

             { al<cl }
             { es sind mehr zeichen vorhanden, als bentigt werden }
             { gerade anzahl? }
             or  ah, ah
             jnz @ungerade

             { gerade, dann ist ein zeichen mehr erlaubt }
             inc al
             cmp al, cl
             jne @not_decodable
             jmp @3

           @ungerade:
             { ungerade, dann zwei mehr erlaubt }
             add al, 2
             cmp al, cl
             jb @not_decodable

           @3:
             mov cl, dh {anzahl der entstehenden zeichen}

           @schleife_zeichen_konvertieren:
             lodsw

             sub al, 32
             jb @not_decodable
             cmp al, 64
             ja @not_decodable

             sub ah, 32
             jb @not_decodable
             cmp ah, 64
             ja @not_decodable
             and ax, 0011111100111111b

  {ein Zeichen aus ax abspeichern, bertrag in dl retten}
             mov dl, ah
             shl al, 2
             shr ah, 4
             or  al, ah
             stosb

             lodsw

             cmp cl, 2
             ja @weiter      {=3}
             je @ein_zeichen {=2}
                             {=1}
             mov al, 32
           @ein_zeichen:
             mov ah, 32

           @weiter:
             sub al, 32
             jb @not_decodable
             cmp al, 64
             ja @not_decodable

             sub ah, 32
             jb @not_decodable
             cmp ah, 64
             ja @not_decodable
             and ax, 0011111100111111b

           { dl, al, ah speichern }
             mov bl, al
             shl bl, 6
             or  ah, bl
             shr al, 2
             shl dl, 4
             or  al, dl
             stosw

             sub cl, 3
             jc  @schleife_fertig
             jnz @schleife_zeichen_konvertieren
           @schleife_fertig:

             xor ax, ax
             mov al, dh
             add ufield.bufpos, ax

           @decodable:
             mov dl, ufield.declen
             mov ufield.declen, dh

             cmp dl, declen_init { kleiner als vorige zeile? }
             je  @firstline
             cmp dh, dl
             jne @decodable_lastline
             jmp @d1

           @firstline:
             mov al, 1
             mov ufield.decopen, al

{$IFDEF OS2} mov ebx, fdestmempos        {$ELSE} mov bx, word ptr fdestmempos     {$ENDIF}
{$IFDEF OS2} mov ufield.deststart, ebx   {$ELSE} mov word ptr ufield.deststart, bx  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov bx, word ptr fdestmempos +2  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov word ptr ufield.DestStart+2, bx{$ENDIF}

           @d1:
             cmp dh, mincodesize { weniger als 30 zeichen data ? }
             jb @decodable_lastline

             xor ah, ah          { ungerade anzahl ? }
             mov al, dh
             mov dl, 3
             div dl       { ax div dl }
             { ah: mod }
             cmp ah, 0

             je @laenge_ok

           @decodable_lastline:
             mov dl, 1
             mov ufield.lastline, dl

           @laenge_ok:
             mov ax, 1
             jmp @exit

           @not_decodable:
             xor ax, ax
           @exit:
{$IFDEF OS2} pop ebx                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop edi                     {$ELSE} pop ds                     {$ENDIF}
{$IFDEF OS2} pop esi                     {$ELSE}                            {$ENDIF}
END;

{***************************************************************************}
{ XX-Decodierung }
{***************************************************************************}

FUNCTION XXDecodAble: Boolean; Assembler;
{dekodiert outstring in outfield, wenn mglich}
ASM
{ Zeiger initalisieren }
{$IFDEF OS2} push esi                     {$ELSE} push ds                    {$ENDIF}
{$IFDEF OS2} push edi                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} push ebx                     {$ELSE}                            {$ENDIF}

{$IFDEF OS2} mov edi, xfield.buf          {$ELSE} les di, xfield.buf         {$ENDIF}
{$IFDEF OS2} movzx eax,  xfield.bufpos    {$ELSE} mov ax, xfield.bufpos      {$ENDIF}
{$IFDEF OS2} add edi, eax                 {$ELSE} add di, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov esi, offset outstring    {$ELSE} mov ax, seg outstring      {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov ds, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov si, offset outstring   {$ENDIF}

             cld
             lodsb
{$IFDEF OS2} xor ecx, ecx                 {$ELSE} xor cx, cx                 {$ENDIF}
             mov cl, al
             dec cl      { cl: Anzahl Textzeichen }

           { Gre testen }
             lodsb
{$IFDEF OS2} xor ebx, ebx                 {$ELSE} xor bx, bx                 {$ENDIF}
           { BH nicht benutzen! }
{$IFDEF OS2} mov bl,  al                  {$ELSE} mov bl, al                 {$ENDIF}
{$IFDEF OS2} add ebx, offset xxfield      {$ELSE} mov al, [bx+offset xxfield]{$ENDIF}
{$IFDEF OS2} mov al,  [ebx]               {$ELSE}                            {$ENDIF}

             cmp al,  255
             je  @not_decodable
             or  al, al
             jnz @0

             cmp cl, 0
             jne @not_decodable

             xor dh, dh
             mov xfield.declen, 0    { sonst Probleme bei firstline }
             jmp @decodable { Zeile mit "+" }

            @0:
             mov ah, xfield.lastline
             or  ah, ah
             jnz @not_decodable

             mov dh, al  { dh: Anzahl der Zeichen, die entstehen sollen }

             { Gre mit voriger Zeile vergleichen }

             mov ah, xfield.declen
             cmp ah, declen_init
             je @line_ok

             cmp dh, ah
             ja @not_decodable

           @line_ok:
             xor ah, ah  { div 3, +1?, *4 }
             shl ax, 2

             mov dl, 3
             div dl       { ax div dl }
             { ah: mod }
             or  ah, ah
             jz @1
             inc al
           @1: { al: Anzahl der dafr bentigten Textzeichen }

             cmp al, cl
             ja @not_decodable { al>cl }
             je @3

             { al<cl }
             { es sind mehr zeichen vorhanden, als bentigt werden }
             { gerade anzahl? }
             or  ah, ah
             jnz @ungerade

             { gerade, dann ist ein zeichen mehr erlaubt }
             inc al
             cmp al, cl
             jne @not_decodable
             jmp @3

           @ungerade:
             { ungerade, dann zwei mehr erlaubt }
             add al, 2
             cmp al, cl
             jb @not_decodable

           @3:
             mov cl, dh {Anzahl d zu entstehenden Zeichen}

           @schleife_zeichen_konvertieren:
             lodsw

  { Zeichen 1}
{$IFDEF OS2} movzx ebx,  al              {$ELSE} mov bl, al                 {$ENDIF}
{$IFDEF OS2} add ebx, offset xxfield     {$ELSE} mov al, [bx+offset xxfield]{$ENDIF}
{$IFDEF OS2} mov al, [ebx]               {$ELSE}                            {$ENDIF}
             cmp al, 255
             je @not_decodable

  { Zeichen 2}
{$IFDEF OS2} movzx ebx,  ah              {$ELSE} mov bl, ah                 {$ENDIF}
{$IFDEF OS2} add ebx, offset xxfield     {$ELSE} mov ah, [bx+offset xxfield]{$ENDIF}
{$IFDEF OS2} mov ah, [ebx]               {$ELSE}                            {$ENDIF}
             cmp ah, 255
             je @not_decodable

  {ein Zeichen aus ax abspeichern, bertrag in dl retten}
             mov dl, ah
             shl al, 2
             shr ah, 4
             or  al, ah
             stosb

             lodsw
             cmp cl, 2
             ja @weiter      {=3}
             je @ein_zeichen {=2}
                             {=1}
             mov al, 43
            @ein_zeichen:
             mov ah, 43

            @weiter:
  { Zeichen 3}
{$IFDEF OS2} movzx ebx,  al              {$ELSE} mov bl, al                 {$ENDIF}
{$IFDEF OS2} add ebx, offset xxfield     {$ELSE} mov al, [bx+offset xxfield]{$ENDIF}
{$IFDEF OS2} mov al, [ebx]               {$ELSE}                            {$ENDIF}
             cmp al, 255
             je @not_decodable

  { Zeichen 4}
{$IFDEF OS2} movzx ebx,  ah              {$ELSE} mov bl, ah                 {$ENDIF}
{$IFDEF OS2} add ebx, offset xxfield     {$ELSE} mov ah, [bx+offset xxfield]{$ENDIF}
{$IFDEF OS2} mov ah, [ebx]               {$ELSE}                            {$ENDIF}
             cmp ah, 255
             je @not_decodable

           { dl, al, ah speichern }
             mov bl, al
             shl bl, 6
             or  ah, bl
             shr al, 2
             shl dl, 4
             or  al, dl
             stosw

             sub cl, 3
             jc  @schleife_fertig
             jnz @schleife_zeichen_konvertieren
           @schleife_fertig:

             xor ax, ax
             mov al, dh
             add xfield.bufpos, ax

           @decodable:
             mov dl, xfield.declen
             mov xfield.declen, dh

             cmp dl, declen_init { kleiner als vorige zeile? }
             je  @firstline
             cmp dh, dl
             jne @decodable_lastline
             jmp @d1

           @firstline:
             mov al, 1
             mov xfield.decopen, al

{$IFDEF OS2} mov ebx, fdestmempos        {$ELSE} mov bx, word ptr fdestmempos     {$ENDIF}
{$IFDEF OS2} mov xfield.DestStart, ebx     {$ELSE} mov word ptr xfield.DestStart, bx  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov bx, word ptr fdestmempos +2  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov word ptr xfield.DestStart+2, bx{$ENDIF}

           @d1:
             cmp dh, mincodesize { weniger als 30 zeichen data ? }
             jb  @decodable_lastline

             xor ah, ah          { ungerade anzahl ? }
             mov al, dh
             mov dl, 3
             div dl       { ax div dl }
             { ah: mod }
             cmp ah, 0

             je @laenge_ok

           @decodable_lastline:
             mov dl, 1
             mov xfield.lastline, dl

           @laenge_ok:
             mov ax, 1
             jmp @exit

           @not_decodable:
             xor ax, ax
           @exit:
{$IFDEF OS2} pop ebx                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop edi                     {$ELSE} pop ds                     {$ENDIF}
{$IFDEF OS2} pop esi                     {$ELSE}                            {$ENDIF}
END;

{***************************************************************************}
{ Mime-Decodierung }
{***************************************************************************}

FUNCTION Base64DecodAble: Boolean; Assembler;
{dekodiert outstring in outfield, wenn mglich}
ASM
{ Zeiger initalisieren }
{$IFDEF OS2} push esi                     {$ELSE} push ds                    {$ENDIF}
{$IFDEF OS2} push edi                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} push ebx                     {$ELSE}                            {$ENDIF}

             mov ah, bfield.lastline
             or  ah, ah
             jnz @not_decodable

{$IFDEF OS2} mov edi, bfield.buf          {$ELSE} les di, bfield.buf         {$ENDIF}
{$IFDEF OS2} movzx eax,  bfield.bufpos    {$ELSE} mov ax, bfield.bufpos      {$ENDIF}
{$IFDEF OS2} add edi, eax                 {$ELSE} add di, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov esi, offset outstring    {$ELSE} mov ax, seg outstring      {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov ds, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov si, offset outstring   {$ENDIF}

             cld
             lodsb
             mov cl, al
               { cl: Anzahl Textzeichen }

             xor ah, ah
             shr al, 2
             mov dl, 3
             mul dl

             mov ah, cl
             and ah, 11b { rest ? }

             { al: Zeichen, die entstehen sollen }
             { ah: Rest }

             mov dl, bfield.declen
             cmp dl, declen_init
             je @first_line

             cmp al, dl
             ja @not_decodable
             jmp @line_ok
             { letzte zeile: ungerade anzahl erlaubt }

           @first_line:
             or  ah, ah
             jnz @not_decodable

             cmp al, mincodesize
             jb @not_decodable

           @line_ok:
             xor ch, ch  { ch: Endmarker gefunden ? }
             xor dh, dh  { dh: Anzahl d. dekodierten Zeichen }
             xor bh, bh  { BH nicht benutzen! }

           { jeweils vier Zeichen in drei Byte dekodieren: }
           @schleife_zeichen_konvertieren:
             lodsw

{$IFDEF OS2} movzx ebx, al                {$ELSE} mov bl, al                 {$ENDIF}
{$IFDEF OS2} add ebx, offset base64       {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov al,  [ebx]               {$ELSE} mov al, [bx+offset base64] {$ENDIF}
             cmp al, 128
             ja @not_decodable

{$IFDEF OS2} movzx ebx, ah                {$ELSE} mov bl, ah                 {$ENDIF}
{$IFDEF OS2} add ebx, offset base64       {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov ah,  [ebx]               {$ELSE} mov ah, [bx+offset base64] {$ENDIF}
             cmp ah, 128
             jb @zeichen2_okay
             ja @not_decodable

             mov ch, 1
             xor ah, ah
             {dec dh} { da gleich inc }
             jmp @zeilenende

           @zeichen2_okay:
             inc dh

           { 1. Zeichen abspeichern, liegt in ax vor, bertrag in dl retten!}
             mov dl, ah
             shl al, 2
             shr ah, 4
             or  al, ah
             stosb

             lodsw

{$IFDEF OS2} movzx ebx, al                {$ELSE} mov bl, al                 {$ENDIF}
{$IFDEF OS2} add ebx, offset base64       {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov al,  [ebx]               {$ELSE} mov al, [bx+offset base64] {$ENDIF}
             cmp al, 128
             jb @zeichen3_okay
             ja @not_decodable

             mov ch, 1
             xor al, al
             {dec dh}
             jmp @zeilenende

           @zeichen3_okay:
             inc dh

{$IFDEF OS2} movzx ebx, ah                {$ELSE} mov bl, ah                 {$ENDIF}
{$IFDEF OS2} add ebx, offset base64       {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov ah,  [ebx]               {$ELSE} mov ah, [bx+offset base64] {$ENDIF}
             cmp ah, 128
             jb @zeichen4_okay
             ja @not_decodable

             mov ch, 1
             xor ah, ah
             dec dh

           @zeichen4_okay:
             inc dh

           { dl, al, ah speichern }
             mov bl, al
             shl bl, 6
             or  ah, bl
             shr al, 2
             shl dl, 4
             or  al, dl
             stosw

             or  ch, ch
             jnz @zeilenende

             sub cl, 4
             cmp cl, 4
             jae @schleife_zeichen_konvertieren

           @zeilenende:
             cmp bfield.declen, declen_init
             jne @01
             cmp ch, 1
             je @not_decodable { erste und letztze Zeile bei Base64=Mll }

           @01:
             or  dh, dh
             jz @not_decodable

             xor ah, ah
             mov al, dh
             add bfield.bufpos, ax

             cmp cl, 1
             jne @decodable
             lodsb
             cmp al, '='
             jne @decodable
             mov ch, 1

           @decodable:
             mov dl, bfield.declen
             mov bfield.declen, dh

             cmp dl, declen_init { kleiner als vorige zeile? }
             je @firstline
             cmp dh, dl
             jne @lastline
             jmp @d1

           @firstline:
             mov al, 1
             mov bfield.decopen, al

{$IFDEF OS2} mov ebx, fdestmempos        {$ELSE} mov bx, word ptr fdestmempos     {$ENDIF}
{$IFDEF OS2} mov bfield.DestStart, ebx     {$ELSE} mov word ptr bfield.DestStart, bx  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov bx, word ptr fdestmempos +2  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov word ptr bfield.DestStart+2, bx{$ENDIF}

           @d1:
             cmp ch, 1
             je  @lastline
             cmp dh, mincodesize { weniger als 30 zeichen data ? }
             jb  @lastline

             xor ah, ah          { ungerade anzahl ? }
             mov al, dh
             mov dl, 3
             div dl       { ax div dl }
             { ah: mod }
             cmp ah, 0

             je @laenge_ok

           @lastline:
             mov dl, 1
             mov bfield.lastline, dl

           @laenge_ok:
             mov ax, 1
             jmp @exit

           @not_decodable:
             xor ax, ax

           @exit:
{$IFDEF OS2} pop ebx                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop edi                     {$ELSE} pop ds                     {$ENDIF}
{$IFDEF OS2} pop esi                     {$ELSE}                            {$ENDIF}
END;

{***************************************************************************}
{ BinHex - Dekodierung: 1:1-bersetzung }
{***************************************************************************}

FUNCTION BinHexDecodAble: Boolean; Assembler;
{dekodiert outstring in outfield, wenn mglich}
ASM
{ Zeiger initalisieren }
{$IFDEF OS2} push esi                     {$ELSE} push ds                    {$ENDIF}
{$IFDEF OS2} push edi                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} push ebx                     {$ELSE}                            {$ENDIF}

             mov ah, hfield.lastline
             or  ah, ah
             jnz @not_decodable

{$IFDEF OS2} mov edi, hfield.buf          {$ELSE} les di, hfield.buf         {$ENDIF}
{$IFDEF OS2} movzx eax,  hfield.bufpos    {$ELSE} mov ax, hfield.bufpos      {$ENDIF}
{$IFDEF OS2} add edi, eax                 {$ELSE} add di, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov esi, offset outstring    {$ELSE} mov ax, seg outstring      {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov ds, ax                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov si, offset outstring   {$ENDIF}
           { Zeilenlnge berprfen }
             cld
             lodsb
{$IFDEF OS2} movzx ecx, al                {$ELSE} xor ch, ch                 {$ENDIF}
{$IFDEF OS2}                              {$ELSE} mov cl, al                 {$ENDIF}
           { -> cl: Anzahl Textzeichen }
             cmp cl, 65
             je @zeilenlaenge_ok
             ja @not_decodable
             cmp cl, 1              { zwei Zeichen mssen es schon sein! }
             je @not_decodable

           @zeilenlaenge_ok:
             mov dx, cx
{$IFDEF OS2}                              {$ELSE} xor bh, bh                 {$ENDIF}

           @schleife_zeichen_konvertieren:
             lodsb

{$IFDEF OS2} movzx ebx, al                {$ELSE} mov bl, al                 {$ENDIF}
{$IFDEF OS2} add ebx, offset BinHexField  {$ELSE}                            {$ENDIF}
{$IFDEF OS2} mov al, [ebx]                {$ELSE} mov al, [bx+offset BinHexField]{$ENDIF}

             cmp al, 128
             ja @not_decodable
             jb @zeichen_ok
             { Ende-Zeichen gefunden: }
             cmp cx, 1
             je @zeichen_ok
             cmp hfield.declen, declen_init
             jne @not_decodable

           @zeichen_ok:
             stosb

             dec cx
             jnz @schleife_zeichen_konvertieren

             cmp dl, 64         { entweder 64 Zeichen oder ':' am Ende }
             je @decodable
             cmp al, 128
             jne @not_decodable
             mov hfield.lastline, 1

           @decodable:
             add hfield.bufpos, dx
             mov ax, 1
             cmp hfield.decopen, al
             je @nicht_firstline

             mov hfield.decopen, al

{$IFDEF OS2} mov ebx, fdestmempos        {$ELSE} mov bx, word ptr fdestmempos       {$ENDIF}
{$IFDEF OS2} mov hfield.DestStart, ebx   {$ELSE} mov word ptr hfield.DestStart, bx  {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov bx, word ptr fdestmempos +2    {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov word ptr hfield.DestStart+2, bx{$ENDIF}

           @nicht_firstline:
             mov hfield.declen, dl
             jmp @exit

           @not_decodable:
{$IFDEF OS2} xor ax, ax                   {$ELSE} xor ax, ax                 {$ENDIF}

           @exit:
{$IFDEF OS2} pop ebx                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop edi                      {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop esi                      {$ELSE} pop ds                     {$ENDIF}
END;

{***************************************************************************}
{ schneller Stringvergleich }
{***************************************************************************}

FUNCTION OutStringStart_NoCaseSensity(Var Buffer): Boolean; Assembler;
ASM
{$IFDEF OS2} push esi                    {$ELSE} push ds                    {$ENDIF}
{$IFDEF OS2} push edi                    {$ELSE}                            {$ENDIF}
{$IFDEF OS2} push ebx                     {$ELSE}                            {$ENDIF}

{$IFDEF OS2} mov edi, Buffer             {$ELSE} les di, Buffer             {$ENDIF}
{$IFDEF OS2} mov esi, offset OutString   {$ELSE} mov ax, seg outstring      {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov ds, ax                 {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov si, offset outstring   {$ENDIF}
{$IFDEF OS2} mov ebx, offset UpCaseField {$ELSE} mov bx, offset UpCaseField {$ENDIF}

             cld
{$IFDEF OS2}                             {$ELSE}                            {$ENDIF}
           { Lnge ok ? }
{$IFDEF OS2} mov cl, [edi]               {$ELSE} mov cl, [di]                           {$ENDIF}
             lodsb
             cmp al, cl
             jb @false

           @schleife:
{$IFDEF OS2} inc edi                     {$ELSE} inc di                     {$ENDIF}
{$IFDEF OS2} mov ah, [edi]               {$ELSE} mov ah, [di]               {$ENDIF}
             lodsb
             xlat
             cmp ah, al
             jne @false

             dec cl
             jnz @schleife

           @true:
             mov ax, 1
             jmp @exit

           @false:
             xor ax, ax

           @exit:
{$IFDEF OS2} pop ebx                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop edi                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop esi                     {$ELSE} pop ds                     {$ENDIF}
END;

{***************************************************************************}
{ schneller Stringvergleich }
{***************************************************************************}

FUNCTION OutStringStart_CaseSensity(Var Buffer): Boolean; Assembler;
ASM
{$IFDEF OS2} push esi                    {$ELSE} push ds                    {$ENDIF}
{$IFDEF OS2} push edi                    {$ELSE}                            {$ENDIF}

{$IFDEF OS2} mov edi, Buffer             {$ELSE} les di, Buffer             {$ENDIF}
{$IFDEF OS2} mov esi, offset OutString   {$ELSE} mov ax, seg outstring      {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov ds, ax                 {$ENDIF}
{$IFDEF OS2}                             {$ELSE} mov si, offset outstring   {$ENDIF}

             cld
{$IFDEF OS2}                             {$ELSE}                            {$ENDIF}
           { Lnge ok ? }
{$IFDEF OS2} mov cl, [edi]               {$ELSE} mov cl, [di]                           {$ENDIF}
             lodsb
             cmp al, cl
             jb @false

           @schleife:
{$IFDEF OS2} inc edi                     {$ELSE} inc di                     {$ENDIF}
{$IFDEF OS2} mov ah, [edi]               {$ELSE} mov ah, [di]               {$ENDIF}
             lodsb
             cmp ah, al
             jne @false

             dec cl
             jnz @schleife

           @true:
             mov ax, 1
             jmp @exit

           @false:
             xor ax, ax

           @exit:
{$IFDEF OS2} pop edi                     {$ELSE}                            {$ENDIF}
{$IFDEF OS2} pop esi                     {$ELSE} pop ds                     {$ENDIF}
END;

{***************************************************************************}
{ Vorbereitung der Decodierung }
{***************************************************************************}

{$IFDEF Debug_Dec}
var tmp:text;
    tmpbuf:array[1..8192] of char;
{$ENDIF}

  PROCEDURE ResetField(Var x: tDecodeArea);
  BEGIN
    x.Buf        :=@x.OrgBuf^;
    x.BufPos     :=0;
    x.DecLen     :=declen_init;
    x.LastLine   :=False;
    x.DecOpen    :=False;
    x.DestStart  :=DestEndUnknown;
    x.DestEnd    :=DestEndUnknown;
  END;

PROCEDURE CloseOutFile(Flag: Byte);
{ evtl. geffnete Datei schlieen }
VAR S1 : S80;
BEGIN
  If OutIDX.Typ<>opNone Then Begin
    OutPos:=nField^.BufPos;

    If nField^.LastLine Then Flag:=ifFEnd;
      { wenn letzte Zeile vor Opennewfile endmarker enthielt, geht verloren }
    OutIDX.Flag:=OutIDX.Flag Or Flag;
    Inc(OutIDX.dSize, OutPos-DiffPos);

    If ( (Not PGPBlockDetected) Or (OutIDX.Typ<>opBase64) ) And
       ( (OutIDX.Typ<>opUUorXX) Or (nField^.DecOpen) ) And
       ( (OutIDX.dSize>0) Or (OutIDX.Flag And ifFEnd<>0) ) Then Begin

{$ifdef Debug_Dec}
      writeln(tmp, 'entry stored, type: '+codename(outidx.typ));
{$endif}

      Inc(DatFileSize, OutIDX.dSize); { neue Daten hinzufgen }
      Inc(DatKeepSize, OutIDX.dSize);
      OutIDX.BnkMode:=bnkNo;

      If (((ThisMsgModify=mmShorten) Or (ThisMsgModify=mmMayDelete))
         And (nField^.DestStart<>DestEndUnknown))
{$IFDEF Binkley}
        Or (BnkThisMail And (OutIDX.dSize>=BnkMinCodeSize)) And FDestOpen
{$ENDIF}
         Then Begin


        If Flag=ifFEnd Then nField^.DestEnd:=FDestFilePos+DestPos
        Else

          If nField^.DestEnd=DestEndUnknown Then
            nField^.DestEnd:=FDestMemPos; { aktuelle Position }

{$IFDEF Binkley}
        If (BnkThisMail And (OutIDX.dSize>=BnkMinCodeSize)) Then Begin
          OutIDX.BnkMode:=bnkDecode;
          BnkModifyMail;
          xMemReplaceFile(FDest, nField^.DestStart, nField^.DestEnd,
                                 BnkMessageBuf^, BnkMessageSize,
                                 DestField^, DestPos);
        End Else Begin
{$ENDIF}
{$IFDEF German}
          S1:='[... '+verName+' hat kodierte '+CodeName(OutIDX.Typ)
            +' Daten entfernt ...]'#13#10;
{$ELSE}
          S1:='[... '+verName+' has removed coded '+CodeName(OutIDX.Typ)
            +' data ...]'#13#10;
{$ENDIF}
          If (PKTType=ptPKT) Or (PKTType=ptMSG) Then Dec(S1[0]);

          xMemReplaceFile(FDest, nField^.DestStart, nField^.DestEnd,
                                 S1[1], Ord(S1[0]),
                                 DestField^, DestPos);
{$IFDEF Binkley}
        End;
{$ENDIF}

        If Result<>0 Then ErrorAbort:=eaDatFileWrite;
        FDestFilePos:=xMemFilePos(FDest);
        FDestMemPos:=FDestFilePos+DestPos;

        ThisMsgIsModified:=True;
      End;

      AddIDX(OutIDX);
      Inc(stFileDecode, OutIDX.dSize);
      Inc(stAllFDecode, OutIDX.dSize);
      ThisMsgData:=True;
    End Else Begin
      xMemTruncFile(DatFile, DatFileSize, OutField^, OutPos);
      If Result<>0 Then ErrorAbort:=eaDatFileRead;
    End;

    OutIDX.Typ:=opNone;
    OutIDX.FileTime:=noFileTime;
    ResetField(nField^);
  End;
  FirstNotDecodedLine_Pos:=DestEndUnknown;
END;

{***************************************************************************}
{ Dekodierung }
{***************************************************************************}

PROCEDURE deDecode;

  PROCEDURE VerifyMem;
  { prft, ob noch platz in nfield vorhanden }
  BEGIN
    If nField^.BufPos>=BlockSize Then Begin
      xMemAppendFile(DatFile, OutField^);
      If Result<>0 Then
        If lfnFreeDiskSpace(DBPath[1])<BlockSize
          Then ErrorAbort:=eaDiskFull
          Else ErrorAbort:=eaDatFileWrite;

      Inc(OutIDX.dSize, BlockSize-DiffPos);

      Move(OutField^[BlockSize], OutField^, nField^.BufPos-BlockSize);
      Dec(nField^.BufPos, BlockSize);
      DiffPos:=0;
    End;

    nField^.DestEnd:=DestEndUnknown;
  END;

  PROCEDURE OpenNewFile(Flag, Typ: Byte; Buf: pDecodeArea);
  BEGIN
    If (OutIDX.Typ<>opNone) And (OutIDX.Typ<>opUUorXX) Then
      CloseOutFile(ifNone);

    OutIDX.Typ:=Typ;    { Ausgang geffnet }
    OutIDX.Flag:=Flag;
    OutIDX.Link:=ilUnlinked;
    OutIDX.dStart:=DatFileSize;
    OutIDX.dSize:=0;
    FirstNotDecodedLine_Pos:=DestEndUnknown;

    If (Typ=opBinHex) And (Flag=ifNone) Then
      If hField.Buf^[0]=128 Then OutIDX.Flag:=ifFBegin
      Else OutIDX.Flag:=ifNone;

    DiffPos:=OutPos;
    nField:=Buf;
    Move(nField^.Buf^, OutField^[OutPos], nField^.BufPos);
    Inc(OutPos, nField^.BufPos);
    nField^.Buf:=OutField;
    nField^.BufPos:=OutPos;
    nField^.DestEnd:=DestEndUnknown;

    If Typ<>opUU       Then ResetField(uField);
    If Typ<>opXX       Then ResetField(xField);
    If Typ<>opBase64   Then ResetField(bField);
    If Typ<>opBinHex   Then ResetField(hField);
  END;

  FUNCTION GetUUFName: Boolean;
  VAR i : Byte;
  BEGIN
  { format is 'begin nnn filename' }
  { also:            nnnn }
    i:=6;
    While (OutString[i]=' ') And (i<=Length(OutString)) Do Inc(i);
    While (OutString[i]<>' ') And (i<=Length(OutString)) Do Inc(i);
    While (OutString[i]=' ') And (i<=Length(OutString)) Do Inc(i);

    If ((i=11) Or (i=12)) And (Length(OutString)>i) Then Begin
      OutIDX.rFileName.Name:=fnName(Copy(OutString,i,255));
      OutIDX.rFileName.Long:=True;
      GetUUFName:=True;
    End Else GetUUFName:=False;
  END;

  FUNCTION GetUUSection: Boolean;
  VAR i, j     : Byte;
      cP, cPof : Word;
      fname    : String;
  { format is 'section x[/y] [of [encoder] of] file [filename] [remarks]' }
  BEGIN
    i:=9;
    While (OutString[i] In ['0'..'9']) And (i<=Length(OutString)) Do
      Inc(i);

    cP:=Value(Copy(OutString,9,i-9));

    If (i<=Length(OutString)) And (OutString[i]='/') Then Begin
      j:=i+1;
      While (OutString[j] In ['0'..'9']) And (i<=Length(OutString)) Do
        Inc(j);
      cPof:=Value(Copy(OutString,i+1,j-i-1));
    End Else cPof:=isUnassigned;

    j:=Pos('file', OutString);

    If (cP<>0) And (j<>0) And (Length(OutString)>j+8) Then Begin
      CloseOutFile(ifNone);

      fname:=Copy(OutString, j+8, 255);
      j:=2;
      While (j<=Length(fName)) And (fName[j]<>' ') Do
        Inc(j);
      fName[0]:=Chr(j-1);
      OutIDX.rFileName.Name:=fnName(fname);
      OutIDX.rFileName.Long:=True;

      OutIDX.cPart:=cP;
      OutIDX.cPartOf:=cPof;
      OutIDX.PartStr:=OutString;
      Delete(OutIDX.PartStr, 9, i-9);
      Insert('x', OutIDX.PartStr, 9);

      GetUUSection:=True;
    End Else GetUUSection:=False;
  END;

  PROCEDURE GetMimeFName;
  VAR i, j: Byte;
  BEGIN
    i:=Pos('name=', OutString);
    If i<>0 Then Begin
      Inc(i, 5);
      If Copy(OutString,i,1)='"' Then Inc(i);

      j:=i;
      While (j<=Length(OutString)) And (OutString[j]<>'"') Do
        Inc(j);

      If i<>j Then Begin
        OutIDX.rFileName.Name:=fnName(Copy(OutString, i, j-i));
        OutIDX.rFileName.Long:=True;
      End;
    End;
  END;

  FUNCTION IsFidoSplit: Boolean;
  VAR cS, cO : Word;
      Res    : Boolean;
      i,j    : Byte;
  BEGIN
    Res:=False;

    If (Length(OutString)>=63) And (Copy(OutString, Length(OutString)-10, 11)=
     '+++++++++++') Then Begin

      i:=Length(OutString);
      While OutString[i]='+' Do
        Dec(i);
      While (OutString[i]=' ') And (i>1) Do
        Dec(i);

      j:=i;
      While (OutString[i]<>'/') And (i>1) Do
        Dec(i);


      cO:=Value(Copy(OutString,i,j-i+1));
      cS:=Value(Copy(OutString,i-2,2));

      If (cS<>0) And (cO<>0) And (i>=49) Then Begin
        CloseOutFile(ifNone);

        OutIDX.cSplit:=cS;
        OutIDX.cSplitOf:=cO;
        OutIDX.SplitStr:=Copy(OutString, 9, 35);
        Res:=True;
      End;
    End;

    IsFidoSplit:=Res;
  END;

CONST UUXXBegin   : String[06] = 'begin ';
      UUXXEnd     : String[03] = 'end';
      UUXXSection : String[08] = 'section ';
      UUXXTime    : String[09] = 'filetime ';
      Base64Begin : String[33] = 'CONTENT-TRANSFER-ENCODING: BASE64';
      Base64End   : String[05] = '-----';
      MimeBegin   : String[08] = 'CONTENT-';
      BinHexBegin : String[45] = '(This file must be converted with BinHex 4.0)';
      PGPBegin    : String[14] = '-----BEGIN PGP'{ PUBLIC KEY BLOCK-----'};
      PGPSign     : String[21] = '-----BEGIN PGP SIGNED';
      PGPEnd      : String[12] = '-----END PGP';
      FidoSplit   : String[08] = #1'SPLIT: ';
      FidoSeenBy  : String[09] = 'SEEN-BY: ';
VAR Done: Boolean;
BEGIN
{$IFDEF Debug_Dec}
  writeln(tmp, FDestFilePos+DestPos:10, outpos:5, ufield.bufpos:5, xfield.bufpos:5, bfield.bufpos:5,
  hfield.bufpos:5, outidx.typ:5, outstring);
{$ENDIF}

  { wenn offen: weiter dekodieren }
  If OutIDX.Typ<>opNone Then Begin
    Done:=False;

    If OutIDX.Typ=opUU Then Begin
      If UUDecodAble Then Begin
        VerifyMem;
        Done:=True;
      End Else If OutStringStart_CaseSensity(UUXXEnd) Then CloseOutFile(ifFEnd);

    End Else If OutIDX.Typ=opXX Then Begin
      If XXDecodAble Then Begin
        VerifyMem;
        Done:=True;
      End Else If OutStringStart_CaseSensity(UUXXEnd) Then CloseOutFile(ifFEnd);

    End Else If OutIDX.Typ=opUUorXX Then Begin

      If UUDecodAble Then Begin
        OpenNewFile(ifFBegin, opUU, @uField);
        Done:=True;
      End Else If XXDecodable Then Begin
        OpenNewFile(ifFBegin, opXX, @xField);
        Done:=True;
      End Else If OutStringStart_CaseSensity(UUXXEnd) Then Begin
        OpenNewFile(ifFBegin, opUUorXX, @uField);
        CloseOutFile(ifFEnd);
      End;

    End Else If OutIDX.Typ=opBase64 Then Begin
      If Base64DecodAble Then Begin
        VerifyMem;
        Done:=True;
        If bField.LastLine Then CloseOutFile(ifFEnd);
      End Else If OutStringStart_CaseSensity(Base64End) Then CloseOutFile(ifFEnd);

    End Else If OutIDX.Typ=opBinHex Then Begin
      If BinHexDecodAble Then Begin
        VerifyMem;
        Done:=True;
        If hField.Lastline Then CloseOutFile(ifFEnd);
      End;

    End;


    If Done Then {Begin
      FirstNotDecodedLine_Pos:=DestEndUnknown;}
      Exit{;
    End} Else FirstNotDecodedLine_Pos:=FDestMemPos;
{$ifdef Debug_Dec}
    If nField^.DestEnd=DestEndUnknown Then
      writeln(tmp, 'code end set to last line');
{$endif}
    If nField^.DestEnd=DestEndUnknown Then nField^.DestEnd:=FDestMemPos;
  End;

  { Startmarkierung suchen }
  If OutStringStart_CaseSensity(UUXXBegin) And GetUUFName Then Begin
    If (OutIDX.Typ<>opNone) Then CloseOutFile(ifNone);
    OutIDX.Typ:=opUUorXX; { noch nicht mit OpenNewFile geffnet ! }
    nField:=@uField;
  End Else If OutStringStart_CaseSensity(UUXXSection) And GetUUSection Then Begin
    { Funktion bereits erledigt }
  End Else If OutStringStart_CaseSensity(UUXXTime) Then Begin
    OutIDX.FileTime:=Value(Copy(OutString,10,255));
  End Else If OutStringStart_NoCaseSensity(MimeBegin) Then Begin
    If OutStringStart_NoCaseSensity(Base64Begin) Then
      OpenNewFile(ifFBegin, opBase64, @bField)
    Else Begin
      MimeDetected:=True;
      MimeTextLevel:=mtlInit;
      GetMimeFName;
    End;
  End Else If OutStringStart_CaseSensity(BinHexBegin) Then
    OpenNewFile(ifFBegin, opBinHex, @hField)
  Else If OutStringStart_NoCaseSensity(PGPBegin) And
     (Not OutStringStart_NoCaseSensity(PGPSign)) Then PGPBlockDetected:=True
  Else If PGPBlockDetected And OutStringStart_NoCaseSensity(PGPEnd)
     Then PGPBlockDetected:=False
  Else If OutStringStart_CaseSensity(FidoSplit) And IsFidoSplit Then Begin
  End

  Else Begin

  { andere dekodierbare Zeilen? }
{ n zeichen dekodierbar:
  a) mll, mu nach i zeilen gelscht werden, da
     1) es blockiert richtige dekodierung
     2) es kann flschlich zu echten daten dazugezhlt werden
  b) echte daten
     1) es kommen nochmal daten: sichere dekodierung
     2) es kommt zeilenende: speichern, da split mgl.
}

    If OutStringStart_CaseSensity(FidoSeenBy) Then Exit;
    { in Fido-Netmails koennen lange Seenby-Zeilen faelschlicherweise
      dekodiert werden }

    If OutIDX.Typ<>opUU Then
      If UUDecodAble Then
        If (uField.BufPos-uField.DecLen>=uNormalLineSize)
        Then OpenNewFile(ifNone, opUU, @uField)
        Else
      Else If uField.DecOpen Then
        If OutStringStart_CaseSensity(UUXXEnd) Then Begin
          OpenNewFile(ifNone, opUU, @uField);
          CloseOutFile(ifFEnd);
        End Else ReSetField(uField);

    If OutIDX.Typ<>opXX Then
      If XXDecodAble Then
        If (xField.BufPos-xField.DecLen>=xNormalLineSize)
        Then OpenNewFile(ifNone, opXX, @xField)
        Else
      Else If xField.DecOpen Then
        If OutStringStart_CaseSensity(UUXXEnd) Then Begin
          OpenNewFile(ifNone, opXX, @xField);
          CloseOutFile(ifFEnd);
        End Else ReSetField(xField);

    If OutIDX.Typ<>opBase64 Then
      If Base64DecodAble Then
        If bField.LastLine Then Begin
          OpenNewFile(ifNone, opBase64, @bField);
          CloseOutFile(ifFEnd);
        End Else If (bField.BufPos-bField.DecLen>=bNormalLineSize)
                 Then OpenNewFile(ifNone, opBase64, @bField)
                 Else
      Else If bField.DecOpen Then ReSetField(bField);
      { EndLine kann nicht bercksichtigt werden, da "-----" extrem oft vorkommt }

    If OutIDX.Typ<>opBinhex Then
      If BinHexDecodAble Then
        If (hField.BufPos-hField.DecLen>=hNormalLineSize)
        Then OpenNewFile(ifNone, opBinHex, @hField)
        Else
      Else If hField.DecOpen Then ReSetField(hField);
      { einzelne, letzte BinHex-Zeile kann nicht erkannt werden, da
        "bla:" extrem oft vorkommt }

    If MimeDetected Then GetMimeFName;
  End;
END;

{**************************************************************************}

PROCEDURE deInit;
BEGIN
  PGPBlockDetected:=False;
  MimeDetected:=False;
{ OutIDX ist bereits initalisiert }

  ResetField(uField);
  ResetField(xField);
  ResetField(bField);
  ResetField(hField);

  FirstNotDecodedLine_Pos:=DestEndUnknown;

  deInitialized:=True;

{$IFDEF Debug_Dec}
  writeln(tmp,'SUBJECT: '+outidx.subjstr);
{$ENDIF}
END;

{**************************************************************************}

PROCEDURE deClose;
BEGIN
  If Not deInitialized Then Exit;
  deInitialized:=False;

  CloseOutFile(ifNone);
END;

{**************************************************************************}

PROCEDURE deGetMem;
BEGIN
  New(uField.OrgBuf);
  New(xField.OrgBuf);
  New(bField.OrgBuf);
  New(hField.OrgBuf);

{$IFDEF Debug_Dec}
  assign(tmp, 'c:\pktxcode.tmp');
  rewrite(tmp);
{$ENDIF}
END;

{**************************************************************************}

PROCEDURE deFreeMem;
BEGIN
  Dispose(uField.OrgBuf);
  Dispose(xField.OrgBuf);
  Dispose(bField.OrgBuf);
  Dispose(hField.OrgBuf);

{$IFDEF Debug_Dec}
  close(tmp);
{$ENDIF}
END;

{**************************************************************************}

END.

