{ PktXCode - decodes with UUENCODE, XXENCODE, BASE64 or BINHEX encoded files
             out of text files, Fido-PKT-/MSG- or ZConnect files

  Copyright (C) 1996-98 Ulrich Mueller, email: ulrich.mueller@bigfoot.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


  This source code is compatible with:
  - Borland Pascal 7.0 (DOS)
  - Virtual Pascal 1.1 (OS/2)

}

Unit PKTX_LFN; { Dateifunktionen mit Untersttzung langer Dateinamen (OS2/W95)}

{$I SETTINGS.INI}

{ 29.01.97 Erstellt.
  16.02.97 Pfadfunktionen ergnzt
  24.02.97 lfnFExpand ergnzt
  09.07.97 lange Dateinamen werden berprft
}

INTERFACE

{$IFDEF OS2}
Uses OS2DEF, OS2BASE, VPUTILS,
{$ELSE}
Uses
{$ENDIF}
     DOS;

TYPE fnStr             = String[80];
     tFNStr            = Record
                           Name : fnStr;  {String fr Dateiname}
                           Long : Boolean;
                         End;

CONST ofRead           =  0;
      ofRW             =  1;
      ofAppend         = ofRW; { fr Textfiles }
      ofCreate         =  2;

      roFileMode       = 32; { read only, others may read }
      rwFileMode       = 18; { read/write, deny all accesses by others }
      lfnUNo           =  0;
      lfnUYes          =  1;
      lfnUForce        =  2;

      TextBufferSize = 2048;

      lfnUseLongNames   : Byte       = lfnUYes;
      Result            : Word       = 0;

{$IFNDEF OS2}
TYPE lfnSearchRec      = Record
                           Attr          : LongInt;
                           CreationTime,
                           AccessTime,
                           ModifyTime    : Array[1..8] Of Byte;
                           SizeHi,
                           Size{Lo}      : LongInt;
                           Reserved      : Array[1..8] Of Byte;
                           LongName      : Array[0..259] Of Char;
                           ShortName     : Array[0..13] Of Char;
                           Handle        : Word;
                           Name          : tFNStr;
                           OldSR         : SearchRec;
                         End;
{$ELSE}
TYPE lfnSearchRec      = Record
                           Name          : tFNStr;
                           HDir          : ULong;
                           Attr          : Byte;
                           Time          : Longint;
                           Size          : Longint;
                         End;
{$ENDIF}

TYPE lfnTextBuf      = Record
                         F          : File;
                         Buffer     : Array[0..TextBufferSize-1] Of Char;
                         BufPos     : Word;
                         BufFree    : Word;
                       End;

PROCEDURE lfnOpenFile(Var F: File; Name: String; Mode: Byte);
PROCEDURE lfnCloseFile(Var F: File);
PROCEDURE lfnEraseFile(Name: String);
PROCEDURE lfnRenameFile(Name1, Name2: String);
FUNCTION  lfnExistFile(Name: String): Boolean;
PROCEDURE lfnFindFirst(Path: String; Attr: Word; var F: lfnSearchRec);
PROCEDURE lfnFindNext(var F: lfnSearchRec);
PROCEDURE lfnFindClose(var F: lfnSearchRec);
FUNCTION  lfnDriveSupportsLFN(Drive: Char): Boolean;
PROCEDURE lfnShortenFilename(Var Name: fnStr);
PROCEDURE lfnShortenPathName(Var Path: String);
FUNCTION lfnFExpand(Name: String): String;

FUNCTION lfnExistPath(Path: String): Boolean;
PROCEDURE lfnCreatePath(Var Path: String);
FUNCTION lfnExistCreatePath(Var Path: String): Boolean;

FUNCTION DosChar(C: Char): Boolean;
PROCEDURE GetError;

FUNCTION fnFull(Path: String): String;
FUNCTION fnShortPath(Path: String): String;
FUNCTION fnDir(Path: String): String;
FUNCTION fnName(Path: String): fnStr;
FUNCTION fnExt(Name: tFNStr): fnStr;

FUNCTION lfnFreeDiskSpace(D: Char): LongInt;
FUNCTION lfnFileSizeOf(Name: String): LongInt;

PROCEDURE lfnOpenTextFile(Var F: lfnTextBuf; Name: String; Mode: Byte);
FUNCTION lfnTextFilePos(Var F: lfnTextBuf): LongInt;
PROCEDURE lfnWriteTextFile(Var F: lfnTextBuf; Line: String);
PROCEDURE lfnCloseTextFile(Var F: lfnTextBuf);


IMPLEMENTATION

Uses Strings, pktx_scr;

{***************************************************************************}

PROCEDURE GetError;
BEGIN
  Result:=IOResult;
END;

{***************************************************************************}

FUNCTION DosChar(C: Char): Boolean;
BEGIN
  DosChar:=(C In[#33, #35..#41, #45, #48..#57, #64..#90, #94..#96, #123,
   #125, #126, #128..#255])
END;

{***************************************************************************}

PROCEDURE lfnCheckLongFilename(Var Name: fnStr);
CONST illegal : Array [1..6] of Char = '/*?"<>';
VAR i, j      : Byte;
BEGIN
  For j:=1 To 7 Do
    Repeat
      i:=Pos(illegal[j], Name);
      If i<>0 Then Delete(Name,i,1);
    Until i=0;

  j:=Length(Name);
  While j>2 Do Begin
    If Name[j]=':' Then Delete(Name,j,1);
    Dec(j);
  End;
END;

{***************************************************************************}

PROCEDURE lfnShortenFilename(Var Name: fnStr);
{ macht Namen DOS-kompatibal }
{ aus: "Claudia Schiffer 1024.JPG" -> "CLAU1024.JPG"
       "Testbilder"                -> "TESTBILD"
       "Claudia Schiffer 06/96.TestBild 1.BMP" -> "CLAU0696.BMP" }
VAR i            : Byte;
    N            : String;
    E            : String[4];
    Punkt        : Boolean;
BEGIN
  Punkt:=False;
  i:=Length(Name);
  While i>=1 Do Begin
    Name[i]:=Upcase(Name[i]);
    If Not (DosChar(Name[i]) Or ((Name[i]='.') And (Not Punkt) )) Then
      Delete(Name,i,1);
    If Name[i]='.' Then Punkt:=True;
    Dec(i);
  End;

  i:=Length(Name);
  While (i>=1) And (Name[i]<>'.') Do Dec(i);

  If i=0 Then Begin
    N:=Name;
    E:='';
  End Else Begin
    N:=Copy(Name,1,i-1);
    E:=Copy(Name,i+1,3);
    i:=Pos('.', E);
    If i<>0 Then E[0]:=Chr(i-1);
    E:='.'+E;
  End;

  If Length(N)>8 Then Begin
    i:=Length(N);
    While (i>1) And (N[i] In ['0'..'9']) Do
      Dec(i);

    If (i<>Length(N)) And (Length(N)-i<8) Then
      Delete(N, 9-(Length(N)-i), Length(N)-8)
    Else N[0]:=#8;
  End Else If N='' Then N:='UNKNOWN';
  Name:=N+E;

  If Length(Name)=0 Then Name:='UNKNOWN';
END;

{***************************************************************************}

PROCEDURE lfnShortenPathName(Var Path: String);
VAR Name : fnStr;
BEGIN
  If Length(Path)<=3 Then Exit;

  Path:=fnShortPath(Path);

  Name:=fnName(Path);
  Path:=fnDir(Path);

  lfnShortenPathName(Path);
  lfnShortenFilename(Name);

  Path:=fnFull(fnFull(Path)+Name);
END;

{***************************************************************************}

FUNCTION lfnFExpand(Name: String): String;
VAR S: String;
    i: Byte;
BEGIN
  While (Copy(Name,1,1)='.') Or (Copy(Name,1,1)=':') Do
    Delete(Name,1,1);

  S:=FExpand(Name);
  { Upcase wiederherstellen }
  While Pos(':', Name)<>0 Do
    Delete(Name, 1, Pos(':', Name));

  While Pos('.\', Name)<>0 Do
    Delete(Name, 1, Pos('.\', Name));

  If Copy(S,Length(S)-Length(Name)+1,Length(Name)) = UpcaseStr(Name)
   Then Begin
    Dec(S[0], Length(Name));
    S:=S+Name;
  End;

  lfnFExpand:=S;
END;

{***************************************************************************}


FUNCTION lfnDriveSupportsLFN(Drive: Char): Boolean;

{$IFNDEF OS2}
{INT 21 71A0 - Windows95 - LONG FILENAME - GET VOLUME INFORMATION}
VAR Regs : Registers;
    Root : String[4];
    Buf  : Array[0..31] Of Byte;
BEGIN
  Root:=Upcase(Drive)+':\'#0;

  Regs.AX:=$71A0;
  Regs.DS:=Seg(Root[1]);
  Regs.DX:=Ofs(Root[1]);
  Regs.SI:=Ofs(Root[1]);

  Regs.ES:=Seg(Buf);
  Regs.DI:=Ofs(Buf);
  Regs.CX:=32;

  Intr($21, Regs);

  If (Regs.Flags And FCarry=0) And (Regs.BX And 16384<>0) And (Regs.AX<>$7100)
    Then lfnDriveSupportsLFN:=True
    Else lfnDriveSupportsLFN:=False;
END;

{$ELSE}

{ von Timo Maier@2:2476/830.10 }
VAR X: DriveTypes;
    F: File;
BEGIN
  X:=GetDriveType(Upcase(Drive));

  If X=dtHDHPFS Then lfnDriveSupportsLFN:=True
  Else If (X=dtNovellNet) Or (X=dtLAN) Then Begin
    FileMode:=rwFileMode;
    Assign(F, Upcase(Drive)+':\long filename access check.tmp');
    ReWrite(F,1);
    GetError;
    If Result=0 Then Begin
      Close(F);
      GetError;
      Erase(F);
      GetError;

      lfnDriveSupportsLFN:=True;
    End Else lfnDriveSupportsLFN:=False;
  End Else lfnDriveSupportsLFN:=False;
END;

{$ENDIF}

{***************************************************************************}

{ ofRead, ofRW, ofCreate }

PROCEDURE lfnOpenFile(Var F: File; Name: String; Mode: Byte);
VAR lfn : Boolean;

{$IFNDEF OS2}
  PROCEDURE W95(Mode: Byte);
  VAR Handle, Error       : Word;
      LNamePtr, SNamePtr  : Pointer;
      SName               : String[67];
  BEGIN
    Name[Length(Name)+1]:=#0;
    LNamePtr:=@Name[1];
    SNamePtr:=@SName[1];

    ASM
      push ds

      lds si, lnameptr

      mov ax, $716C

      mov cx, Archive
      mov dl, mode
      cmp dl, ofCreate
      je @mCreate

      mov bx, roFileMode
      mov dx, 00001b {open existing file}
      jmp @mWeiter

     @mCreate:
      mov bx, rwFileMode
      mov dx, 10010b {create new file/truncate if exists}

     @mWeiter:
      int $21
      jc @error

      { Datei schlieen, ax: handle }
      mov bx, ax
      mov ah, $3e
      int $21

      jc @error

      { Namen konvertieren }
      mov ax, $7160
      mov cx, $8001
      lds si, lnameptr
      les di, snameptr
      int $21

      jc @error
      mov error, 0
      jmp @exit

     @error:
      mov error, ax

     @exit:
      pop ds
    END;

    If Error=0 Then Begin
      SName[0]:=#67;
      SName[0]:=Chr(Pos(#0, SName));
      If SName[0]=#0 Then Error:=5;
      Assign(F, SName);
    End;

    Result:=Error;
  END;
{$ENDIF}

VAR lName : fnStr;
BEGIN
  Name:=lfnFExpand(Name);

  lfn:=lfnDriveSupportsLFN(Name[1]);
  If (Mode=ofCreate) And ((Not lfn) Or (lfnUseLongNames=lfnUNo)) Then Begin
    lName:=fnName(Name);
    Name:=fnDir(Name);

    lfnShortenFileName(lName);

    Name:=Name+lName;
  End;
  lfnCheckLongFileName(Name);

  If Mode=ofRead Then FileMode:=roFileMode
  Else FileMode:=rwFileMode;

  Assign(F, Name);

{$IFNDEF OS2}
  If lfn Then w95(Mode);

  If (Not lfn) Or (Result=0) Then Begin
{$ENDIF}
    If Mode=ofCreate Then ReWrite(F, 1)
    Else ReSet(F, 1);
    GetError;
{$IFNDEF OS2}
  End;
{$ENDIF}
END;

{***************************************************************************}

PROCEDURE lfnCloseFile(Var F: File);
BEGIN
  Close(F);
  GetError;
END;

{***************************************************************************}

PROCEDURE lfnEraseFile(Name: String);
VAR F: File;

  PROCEDURE EraseW95;
  VAR NamePtr       : Pointer;
      Error         : Word;
  BEGIN
    Name[Length(Name)+1]:=#0;
    NamePtr:=@Name[1];

    ASM
      push ds

      mov ax, $7141
      lds dx, nameptr
      xor si, si
      mov cl, anyfile
      mov ch, 0

      int $21

      jc @error

      mov error, 0
      jmp @exit

     @error:
      mov error, ax

     @exit:
      pop ds
    END;
    Result:=Error;
  END;

BEGIN
{$IFNDEF OS2}
  If (lfnDriveSupportsLFN(Name[1])) Then EraseW95
  Else Begin
{$ENDIF}
    Assign(F, Name);
    Erase(F);
    GetError;
{$IFNDEF OS2}
  End;
{$ENDIF}
END;

{***************************************************************************}

PROCEDURE lfnRenameFile(Name1, Name2: String);
VAR F: File;

  PROCEDURE RenameW95;
  VAR NPtr1, NPtr2  : Pointer;
      Error         : Word;
  BEGIN
    Name1[Length(Name1)+1]:=#0;
    Name2[Length(Name2)+1]:=#0;
    NPtr1:=@Name1[1];
    NPtr2:=@Name2[1];

    ASM
      push ds

      mov ax, $7156
      lds dx, nptr1
      les di, nptr2

      int $21

      jc @error

      mov error, 0
      jmp @exit

     @error:
      mov error, ax

     @exit:
      pop ds
    END;
    Result:=Error;
  END;

BEGIN
  lfnCheckLongFileName(Name2);
{$IFNDEF OS2}
  If (lfnDriveSupportsLFN(Name1[1])) Then RenameW95
  Else Begin
{$ENDIF}
    Assign(F, Name1);
    Rename(F, Name2);
    GetError;
{$IFNDEF OS2}
  End;
{$ENDIF}
END;

{***************************************************************************}

FUNCTION lfnExistFile(Name: String): Boolean;
VAR F: File;
    r: Boolean;
BEGIN
  lfnOpenFile(F, Name, ofRead);

  r:=(Result=0) Or (Result=5);
  If Result=0 Then lfnCloseFile(F);

  lfnExistFile:=r;
END;

{***************************************************************************}

PROCEDURE lfnFindFirst(Path: String; Attr: Word; var F: lfnSearchRec);
{$IFNDEF OS2}
  PROCEDURE FindFirstW95;
  VAR Regs  : Registers;
  BEGIN
    Path:=Path+#0;

    Regs.AX:=$714E;
    Regs.SI:=$0000;
    Regs.CX:=Attr;
    Regs.DS:=Seg(Path[1]);
    Regs.DX:=Ofs(Path[1]);
    Regs.ES:=Seg(F);
    Regs.DI:=Ofs(F);
    MSDOS(Regs);

    F.Handle:=Regs.AX;
    If Regs.Flags And FCarry<>0 Then DosError:=Regs.AX
    Else DosError:=0;

    F.Name.Name:=StrPas(F.LongName);
    F.Name.Long:=True;
  END;

BEGIN
  If (lfnDriveSupportsLFN(Path[1])) Then FindFirstW95
  Else Begin
    FindFirst(Path, Attr, F.OldSR);
    F.Name.Name:=F.OldSR.Name;
    F.Name.Long:=False;
    F.Size:=F.OldSR.Size;
    F.Attr:=F.OldSR.Attr;
  End;
END;
{$ELSE}
VAR SR: SearchRec;
BEGIN
  FindFirst(Path, Attr, SR);
  F.Name.Name:=SR.Name;
  F.Name.Long:=True;
  F.HDir:=SR.HDir;
  F.Attr:=SR.Attr;
  F.Time:=SR.Time;
  F.Size:=SR.SIze;
END;
{$ENDIF}

{***************************************************************************}

PROCEDURE lfnFindNext(var F: lfnSearchRec);
{$IFNDEF OS2}
  PROCEDURE FindNextW95;
  VAR Regs: Registers;
  BEGIN
    Regs.AX:=$714F;
    Regs.SI:=$0000;
    Regs.BX:=F.Handle;
    Regs.ES:=Seg(F);
    Regs.DI:=Ofs(F);
    MSDOS(Regs);

    If Regs.Flags And FCarry<>0 Then DosError:=Regs.AX
    Else DosError:=0;
    F.Name.Name:=StrPas(F.LongName);
  END;

BEGIN
  If F.Name.Long Then FindNextW95
  Else Begin
    FindNext(F.OldSR);
    F.Name.Name:=F.OldSR.Name;
    F.Size:=F.OldSR.Size;
    F.Attr:=F.OldSR.Attr;
  End;
END;
{$ELSE}
VAR SR: SearchRec;
BEGIN
  SR.HDir:=F.HDir;
  FindNext(SR);
  F.Name.Name:=SR.Name;
  F.Name.Long:=True;
  F.HDir:=SR.HDir;
  F.Attr:=SR.Attr;
  F.Time:=SR.Time;
  F.Size:=SR.SIze;
END;
{$ENDIF}

{***************************************************************************}

PROCEDURE lfnFindClose(var F: lfnSearchRec);
{$IFNDEF OS2}
  PROCEDURE FindCloseW95;
  VAR Regs: Registers;
  BEGIN
    Regs.AX:=$71A1;
    Regs.BX:=F.Handle;
    MSDOS(Regs);

    If Regs.Flags And FCarry<>0 Then DosError:=Regs.AX
    Else DosError:=0;
  END;

BEGIN
  If F.Name.Long Then FindCloseW95;
END;
{$ELSE}
VAR SR: SearchRec;
BEGIN
  SR.HDir:=F.HDir;
  FindClose(SR);
END;
{$ENDIF}

{***************************************************************************}
{ Pfad-Funktionen }
{***************************************************************************}

FUNCTION lfnExistPath(Path: String): Boolean;
VAR F   : File;
    Attr: {$IFDEF OS2} LongInt {$ELSE} Word {$ENDIF} ;

{$IFNDEF OS2}
  PROCEDURE GetFAttrW95;
  VAR NamePtr       : Pointer;
      Error, Attr1  : Word;
      Buf           : Array[0..255] Of Char;
  BEGIN
    StrPCopy(Buf,Path);
    NamePtr:=@Buf;

    ASM
      push ds

      lds dx, nameptr
      mov ax, $7143
      xor bl, bl
      int $21

      jc @error

      mov error, 0
      mov attr1, cx
      jmp @exit

     @error:
      mov error, ax

     @exit:
      pop ds
    END;
    Result:=Error;
    DosError:=0;
    Attr:=Attr1;
  END;
{$ENDIF}

BEGIN
  Path:=fnFull(Path)+'.';

{$IFNDEF OS2}
  If (lfnDriveSupportsLFN(Path[1])) Then GetFAttrW95
  Else Begin
{$ENDIF}
    Assign(F, Path);
    GetFAttr(F, Attr);
    GetError;
{$IFNDEF OS2}
  End;
{$ENDIF}

  If (DosError<>0) Or (Result<>0) Or ( Attr And Directory=0 )
   Then lfnExistPath:=False
   Else lfnExistPath:=True;
END;

{***************************************************************************}

PROCEDURE lfnCreatePath(Var Path: String);

  PROCEDURE MakePath(Path: String);
  VAR Name : fnStr;

{$IFNDEF OS2}
    PROCEDURE MakePathW95;
    VAR NamePtr       : Pointer;
        Error         : Word;
        Buf           : Array[0..255] Of Char;
    BEGIN
      StrPCopy(Buf,Path);
      NamePtr:=@Buf;

      ASM
        push ds

        lds dx, nameptr

        mov ax, $7139
        int $21

        jc @error

        mov error, 0
        jmp @exit

       @error:
        mov error, ax

       @exit:
        pop ds
      END;
      Result:=Error;
    END;
{$ENDIF}

  BEGIN
    If Not lfnExistPath(Path) Then Begin
      Path:=fnShortPath(Path);
      If Length(Path)<=3 Then Exit;

      Name:=fnName(Path);
      Path:=fnDir(Path);

      If Length(Path)>3 Then MakePath(Path);

      Path:=Path+Name;

{$IFNDEF OS2}
      If (lfnDriveSupportsLFN(Path[1])) Then MakePathW95
      Else Begin
{$ENDIF}
        MkDir(Path);
        GetError;
{$IFNDEF OS2}
      End;
{$ENDIF}
    End;
  END;

BEGIN
  If (Not lfnDriveSupportsLFN(Path[1])) Then
    lfnShortenPathName(Path);
  lfnCheckLongFileName(Path);

  If Not lfnExistPath(Copy(fnFull(Path),1,3)) Then Begin
    Result:=3;
    Exit;
  End;

  MakePath(Path);
END;

{***************************************************************************}

FUNCTION lfnExistCreatePath(Var Path: String): Boolean;
VAR r : Boolean;
BEGIN
  Path[1]:=Upcase(Path[1]);

  If (Not lfnDriveSupportsLFN(Path[1])) Then
    lfnShortenPathName(Path);
  lfnCheckLongFileName(Path);

  r:=lfnExistPath(Path);
  If Not r Then Begin
    lfnCreatePath(Path);
    r:=lfnExistPath(Path);
  End;

  lfnExistCreatePath:=r;
END;

{***************************************************************************}
{ String-Behandlung }
{***************************************************************************}


FUNCTION fnFull(Path: String): String;
BEGIN
  If Path[Ord(Length(Path))]='\' Then fnFull:=Path
  Else fnFull:=Path+'\';
END;

FUNCTION fnShortPath(Path: String): String;
BEGIN
  If (Length(Path)=3) Or (Path[Ord(Length(Path))]<>'\') Then fnShortPath:=Path
  Else fnShortPath:=Copy(Path,1,Length(Path)-1);
END;


FUNCTION fnDir(Path: String): String;
VAR n : Byte;
BEGIN
  n:=Length(Path);
  While (n>0) And (Path[n]<>'\') Do
    Dec(n);

  Delete(Path,n+1,255);
  fnDir:=Path;
END;

FUNCTION fnName(Path: String): fnStr;
VAR n : Byte;
BEGIN
  n:=Length(Path);
  While (n>0) And (Path[n]<>'\') Do
    Dec(n);

  Delete(Path,1,n);
  fnName:=Path;
END;

FUNCTION fnExt(Name: tFNStr): fnStr;
VAR i : Byte;
BEGIN
  i:=Length(Name.Name);
  While (i>0) And (Name.Name[i]<>'.') Do Dec(i);

  If i=0 Then fnExt:=''
  Else fnExt:=Copy(Name.Name, i+1, 255);
END;

{***************************************************************************}

FUNCTION lfnFreeDiskSpace(D: Char): LongInt;
{$IFNDEF OS2}
function DiskFree(Drive: Byte): Longint; Assembler;
asm
	MOV	DL,drive
	MOV	AH,36H
	INT	21H
{mov bx,dx}
	MOV	DX,AX
	CMP	AX,0FFFFH
	JE	@@1
	MUL	CX
        jc @big
	MUL	BX

        cmp     dx, $7fff
        jbe @@1
       @big:
        mov     dx, $7fff
        mov     ax, $ffff

@@1:
end;
{$ENDIF}
BEGIN
  lfnFreeDiskSpace:=DiskFree(Ord(UpCase(D))-Ord('A')+1);
END;

{***************************************************************************}

FUNCTION lfnFileSizeOf(Name: String): LongInt;
VAR F   : File;
    Out : LongInt;
BEGIN
  lfnOpenFile(F, Name, ofRead);
  If Result<>0 Then Out:=0
  Else Begin
    Out:=FileSize(F);
    lfnCloseFile(F);
  End;

  lfnFileSizeOf:=Out;
END;

{***************************************************************************}

PROCEDURE lfnOpenTextFile(Var F: lfnTextBuf; Name: String; Mode: Byte);
BEGIN
  F.BufPos:=0;
  F.BufFree:=TextBufferSize;
  lfnOpenFile(F.F, Name, Mode);

  If Mode=ofAppend Then Begin
    If (Result=0) Then Begin
      Seek(F.F, FileSize(F.F));
      GetError;
      F.BufFree:=FileSize(F.F) Mod TextBufferSize;
    End Else Begin
      { neu erstellen }
      lfnOpenFile(F.F, Name, ofCreate);
    End;
  End;
END;

FUNCTION lfnTextFilePos(Var F: lfnTextBuf): LongInt;
BEGIN
  lfnTextFilePos:=FilePos(F.F)+F.BufPos;
END;

PROCEDURE lfnWriteTextFile(Var F: lfnTextBuf; Line: String);
VAR Cpy, i: Byte;
BEGIN
  Line:=Line+#13#10;
  i:=0;
  Repeat
    Cpy:=Length(Line)-i;
    If Cpy>F.BufFree Then Cpy:=F.BufFree;

    Move(Line[i+1], F.Buffer[F.BufPos], Cpy);
    Inc(F.BufPos, Cpy);
    Dec(F.BufFree, Cpy);
    Inc(I, Cpy);

    If F.BufFree=0 Then Begin
      BlockWrite(F.F, F.Buffer, F.BufPos);
      GetError;
      F.BufPos:=0;
      F.BufFree:=TextBufferSize;
    End;
  Until i>=Length(Line);
END;

PROCEDURE lfnCloseTextFile(Var F: lfnTextBuf);
BEGIN
  If F.BufPos>0 Then Begin
    BlockWrite(F.F, F.Buffer, F.BufPos);
    GetError;
  End;

  lfnCloseFile(F.F);
END;

{***************************************************************************}

END.
