{ PktXCode - decodes with UUENCODE, XXENCODE, BASE64 or BINHEX encoded files
             out of text files, Fido-PKT-/MSG- or ZConnect files

  Copyright (C) 1996-98 Ulrich Mueller, email: ulrich.mueller@bigfoot.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


  This source code is compatible with:
  - Borland Pascal 7.0 (DOS)
  - Virtual Pascal 1.1 (OS/2)

}

Unit PKTX_MEM;

{$I SETTINGS.INI}

{ 16.02.97 Neuerstellung
  22.02.97 xMemReplaceFile hinzugefgt
  27.02.97 xMemClose: kein Trunc bei xcRandom
  11.07.97 xMemCloseFile: Result:=0
}

{ Methode: Lesen und Schreiben von 32kb-Blcken wird in Zusatzspeicher
  gepuffert. Bei gleicher Prioritt steht jedem Filehandle GesMem/Anzahl
  Blcke zur Verfgung. Es drfen aber mehr Blcke zur Pufferung reserviert
  werden. Wenn weniger Blcke frei sind, als zustehen, dann wird die Datei
  mit den meisten reservierten Blcken geflushed.
  So wird sichergestellt, da jede Datei ihren zustehenen Speicher bekommt
  und trotzdem mehr Speicher benutzt werden kann, als eigentlich zusteht.
}


INTERFACE

USES PKTX_IO, PKTX_SCR, PKTX_LFN;

CONST MinMemSize          =   128;
      MaxMemSize          = 32768; { Maxgre gesamt }
      OptimalMemSize      =  2048;
      AutoMemSize         = OptimalMemSize-1;

      xcWrite             = 0; { puffert lineares schreiben }
      xcRandom            = 1; { puffert ganze Datei, Anfang in Speicher }

      XMemSize            : Word       = AutoMemSize;

{***************************************************************************}

FUNCTION Addr2Block(Addr: LongInt): Word;
FUNCTION Addr2BlockAddr(Addr: LongInt): LongInt;
FUNCTION Block2Addr(Block: Word): LongInt;

PROCEDURE XMemInit;
PROCEDURE XMemClose;
PROCEDURE XMemShowSize;

PROCEDURE xMemOpenFile(Var F: File; Mode: Byte; Var Buffer; Var BPos: Word);
{ initalisiert Cache fr File "F". Im Modus xcWrite wird "Buffer" mit
  Clustergrenze beschrieben }
PROCEDURE xMemAppendFile(Var F: File; Var Buffer);
PROCEDURE xMemTruncFile(Var F: File; FPos: LongInt; Var Buffer; Var BPos: Word);
PROCEDURE xMemReplaceFile(Var F: File; von, bis: LongInt;
                          Var OrgBuf; OrgSize: Word;
                          Var Buffer; Var BPos: Word);
PROCEDURE xMemReadFile(Var F: File; Pos: LongInt; Size: Word; Var Buffer);
PROCEDURE xMemWriteFile(Var F: File; Pos: LongInt; Size: Word; Var Buffer);
PROCEDURE xMemCloseFile(Var F: File; Var Buffer; Size: Word);
FUNCTION xMemFilePos(Var F: File): LongInt;

{***************************************************************************}

IMPLEMENTATION

Uses DOS;

{***************************************************************************}

  FUNCTION Addr2Block(Addr: LongInt): Word;
  BEGIN
    Addr2Block:=Word(Addr Div BlockSize);
  END;

  FUNCTION Addr2BlockAddr(Addr: LongInt): LongInt;
  BEGIN
    Addr2BlockAddr:=LongInt(Addr Div BlockSize)*BlockSize;
  END;

  FUNCTION Block2Addr(Block: Word): LongInt;
  BEGIN
    Block2Addr:=LongInt(Block)*BlockSize;
  END;

CONST xmiNo           = 0;
      xmiAlloc        = 1;
      xmiYes          = 2;
      XMemInitialized : Byte = xmiNo;

VAR XMemCount       : Word; { Anzahl der Blcke }
    XBuffer         : PField; { Puffer, immer allokieren ! }

Procedure XInit;  Forward;
Procedure XClose; Forward;

{$IFNDEF OS2}
{***************************************************************************}
{*** DOS: XMS-Speicher *****************************************************}
{***************************************************************************}

TYPE XMSRegs       = Record
                       AX,
                       BX,
                       DX,
                       SI,
                       Segment : Word
                     End;

VAR XMSPtr         : Pointer;
    XMSHandle      : Word;

{***************************** XMS-Routinen ********************************}

FUNCTION InitXMS: Boolean;
VAR Regs : Registers;
BEGIN
  Regs.AX:=$4300;
  Intr($2F, Regs);
  If Regs.AL=$80 Then Begin
    Regs.AX:=$4310;
    Intr($2F, Regs);
    XMSPtr:=Ptr(Regs.ES, Regs.BX);
    InitXMS:=True;
  End Else InitXMS:=False;
END;

PROCEDURE XMSCall(FktNr: Byte; Var XRegs: XMSRegs);
BEGIN
  Inline ( $8C / $D9 /                           { mov    cx,ds        }
           $51 /                                 { push   cx           }
           $C5 / $BE / $04 / $00 /               { lds    di,[bp+0004] }
           $8A / $66 / $08 /                     { mov    ah,[bp+0008] }
           $8B / $9D / $02 / $00 /               { mov    bx,[di+0002] }
           $8B / $95 / $04 / $00 /               { mov    dx,[di+0004] }
           $8B / $B5 / $06 / $00 /               { mov    si,[di+0006] }
           $8E / $5D / $08 /                     { mov    ds,[di+08]   }
           $8E / $C1 /                           { mov    es,cx        }
           $26 / $FF / $1E / XMSPtr /            { call   es:[XMSPtr]  }
           $8C / $D9 /                           { mov    cx,ds        }
           $C5 / $7E / $04 /                     { lds    di,[bp+04]   }
           $89 / $05 /                           { mov    [di],ax      }
           $89 / $5D / $02 /                     { mov    [di+02],bx   }
           $89 / $55 / $04 /                     { mov    [di+04],dx   }
           $89 / $75 / $06 /                     { mov    [di+06],si   }
           $89 / $4D / $08 /                     { mov    [di+08],cx   }
           $1F                                   { pop    ds           }
        );
END;

PROCEDURE XMSVersion(Var VerNr, RevNr: Word);
VAR XR: XMSRegs;
BEGIN
  XmsCall(0, XR);

  VerNr:=Hi(XR.AX)*100 +(Lo(XR.AX) Shr 4)*10 + (Lo(XR.AX) And 15);
  RevNr:=Hi(XR.BX)*100 +(Lo(XR.BX) Shr 4)*10 + (Lo(XR.BX) And 15);
END;

PROCEDURE XMSFree(Var MaxBl: Word);
VAR Xr : XMSRegs;
BEGIN
  XMSCall(8, Xr);
  MaxBl:=Xr.DX;

  If Lo(XR.BX)<>0 Then MaxBl:=0; { ErrorAbort:=eaXMS; }
END;

FUNCTION XMSGetMem(LenKb: Word; Var Handle: Word): Boolean;
VAR Xr : XMSRegs;
BEGIN
  Xr.DX:=LenKB;
  XMSCall(9, Xr);
  If (XR.AX=$0001) Then Begin
    Handle:=Xr.DX;
    XMSGetMem:=True;
  End Else XMSGetMem:=False;
END;

PROCEDURE XMSFreeMem(Handle: Word);
VAR Xr : XMSRegs;
BEGIN
  Xr.DX:=Handle;
  XmsCall(10, Xr);
  If (XR.AX=0) And (Lo(XR.BX)<>0) Then ErrorAbort:=eaXMS;
END;

PROCEDURE XMSCopy(VonHandle : Word; VonOffset : LongInt;
                  NachHandle: Word; NachOffset: LongInt;
                  LenW      : LongInt);
TYPE EMMS = Record
              LenB    : LongInt;
              SHandle : Integer;
              SOffset : LongInt;
              DHandle : Integer;
              DOffset : LongInt;
            End;
VAR Xr : XMSRegs;
    Mi : EMMS;
BEGIN
  With Mi Do Begin
    LenB:=LenW;
    SHandle:=VonHandle;
    SOffset:=VonOffset;
    DHandle:=NachHandle;
    DOffset:=NachOffset
  End;
  Xr.Si:=Ofs(Mi);
  Xr.Segment:=Seg(Mi);
  XmsCall(11, Xr);
  If (XR.AX=0) And (Lo(XR.BX)<>0) Then ErrorAbort:=eaXMS;
END;

{***************************************************************************}

PROCEDURE XMemRead (Block: Word; Var Buffer);
BEGIN
  XMSCopy(XMSHandle, LongInt(Block)*BlockSize, 0, LongInt(@Buffer),
    BlockSize);
END;

PROCEDURE XMemWrite(Block: Word; Var Buffer);
BEGIN
  XMSCopy(0, LongInt(@Buffer), XMSHandle, LongInt(Block)*BlockSize,
    BlockSize);
END;

{***************************************************************************}

PROCEDURE XMemInit;
VAR  MaxBlock : Word;
    Time      : LongInt;

Function IsWindows : Boolean;
var regs: registers;
{ Routine to determine if Windows is currently running }
begin
  Regs.AX := $4680;    { Win 3.x Standard check }
  Intr($2F, Regs);     { Call Int 2F }
  if Regs.AX <> 0 then { if AX = 0 Win in Real mode }
  begin                { else check For enhanced mode }
    Regs.AX := $1600;  { Win 3.x Enhanced check }
    Intr($2F, Regs);   { Call Int 2F }
    if Regs.AL in [$00,$80,$01,$FF] then { Check returned value }
      IsWindows := False   { Nope not installed }
    else
      IsWindows := True;   { Ya it is }
  end
  else
    IsWindows := True;     { Return True }
end;

FUNCTION IsWin95: Boolean; Assembler;
ASM
  {INT 2F 4A33 - Windows95 - CHECK MS-DOS VERSION 7}
  push ds
  push si
  mov ax, $4a33
  int $2F
  cmp ax, 0
  je @win95
  xor ax, ax
  jmp @exit

 @win95:
  mov ax, 1
 @exit:
  pop si
  pop ds
END;

{$IFDEF German}
CONST Txt1 = 'Benutze bis zu ';
      Txt2 = 'Leere Smartdrive-Cache';
      Txt3 = 'kein Zugriff auf XMS-Speicher mglich';
{$ELSE}
CONST Txt1 = 'Using up to ';
      Txt2 = 'Flushing smartdrive cache';
      Txt3 = 'unable to allocate XMS memory';
{$ENDIF}
BEGIN
  If XMemInitialized<>xmiNo Then Exit;

  If InitXMS Then Begin
    If XMemSize=AutoMemSize Then Begin
      XMemSize:=MaxMemSize;

{      If (IsWin95 And IsWindows)
        Then Begin
          If (JobExec And jobDecodeFiles<>0) And (stAllFSize>1024*1024)
            Then XMemSize:=MaxMemSize
            Else XMemSize:=0
        End Else}
          If IsWindows Or (Lo(DosVersion)>=10) Then XMemSize:=OptimalMemSize;
    End;

    XMSFree(MaxBlock);
    MaxBlock:=(MaxBlock Div BlockSizeK)*BlockSizeK;
    If MaxBlock>XMemSize Then MaxBlock:=XMemSize;

    If MaxBlock>=MinMemSize Then Begin
      If MaxBlock>MaxMemSize Then MaxBlock:=MaxMemSize;
      If XMSGetMem(MaxBlock, XMSHandle) Then Begin
        XMemCount:=MaxBlock Div BlockSizeK;

        Write(Txt1, MaxBlock, ' kb XMS');
        If JobExec And JobDecodeFiles=0 Then WriteLn;

        { SmartDrv flushen }
        OpenWaitWindow(Txt2);
        Time:=GetTimeReal;
        Asm
          mov ax, $4A10
          mov bx, $0001
          int $2f
        End;
        Time:=GetTimeReal-Time;
        CloseWindow;

        Inc(stProcessTime, Time);

        XMemCount:=MaxBlock Div BlockSizeK;
        XMemInitialized:=xmiYes;
      End Else WriteLn(Txt3);
    End;
  End;

  XInit;
END;

{***************************************************************************}

PROCEDURE XMemClose;
BEGIN
  If XMemInitialized=xmiNo Then Exit;

  XClose;

  If XMemInitialized=xmiYes Then XMSFreeMem(XMSHandle);

  XMemInitialized:=xmiNo;
END;

{***************************************************************************}
{***************************************************************************}
{***************************************************************************}

{$ELSE}

{***************************************************************************}
{*** OS2: Hauptspeicher ****************************************************}
{***************************************************************************}

TYPE tBaseMemPtr = Array[0..BlockSize*1024-1] Of Byte;
VAR BaseMemPtr   : ^tBaseMemPtr;

PROCEDURE XMemInit;
{$IFDEF German}
CONST Txt1 = 'Benutze bis zu ';
{$ELSE}
CONST Txt1 = 'Using up to ';
{$ENDIF}
VAR Avail       : LongInt;
BEGIN
  If XMemInitialized<>xmiNo Then Exit;

  If XMemSize=AutoMemSize Then XMemSize:=OptimalMemSize;

  Avail:=MaxAvail;
  Avail:=Avail Div 1024;
  If Avail<=384 Then Exit;

  Dec(Avail, 384);
  If Avail>XMemSize Then Avail:=XMemSize;

  If Avail>=MinMemSize Then Begin
    Avail:=(Avail Div BlockSizeK)*BlockSizeK;
    If Avail>MaxMemSize Then Avail:=MaxMemSize;

    GetMem(BaseMemPtr, Avail*1024);

    Write(Txt1, Avail,' kb RAM');
    If JobExec And JobDecodeFiles=0 Then WriteLn;

    XMemCount:=Avail Div BlockSizeK;
    XMemInitialized:=xmiYes;
  End;

  XInit;
END;

{***************************************************************************}

PROCEDURE XMemClose;
BEGIN
  If XMemInitialized=xmiNo Then Exit;

  XClose;

  If XMemInitialized=xmiYes Then
    FreeMem(BaseMemPtr, LongInt(XMemSize)*BlockSize);

  XMemInitialized:=xmiNo;
END;

{***************************************************************************}

PROCEDURE XMemRead (Block: Word; Var Buffer);
BEGIN
  Move(BaseMemPtr^[LongInt(Block)*BlockSize], Buffer, BlockSize);
END;

PROCEDURE XMemWrite(Block: Word; Var Buffer);
BEGIN
  Move(Buffer, BaseMemPtr^[LongInt(Block)*BlockSize], BlockSize);
END;

{***************************************************************************}

{$ENDIF}

{***************************************************************************}
{***************************************************************************}
{***************************************************************************}

CONST MaxBlocks  = MaxMemSize Div BlockSizeK;
      MaxFiles   = 8;
      unused     = $ffff;

TYPE tFileInfo   = Record
                     Used      : Boolean;
                     F         : ^File;
                     Mode      : Byte;
                     StartBlock: Word;    { erster Block }
                     BCount    : Word;    { Anzahl der benutzten Blcke }
                     FStart    : LongInt; { Position, an die Cache geschr. wird }
                     FEnd      : LongInt; { Position, bis zur der C reicht }
                   End;

     aFileInfo   = Array[0..Maxfiles-1] Of tFileInfo;

     tBlockInfo  = Record
                     Used      : Boolean;
                     Dirty     : Boolean; { ungleich Festplatte }
                     Filled    : Boolean; { mit Daten geladen? }
                     Next      : Word;
                   End;

     aBlockInfo  = Array[0..MaxBlocks-1] Of tBlockInfo;

VAR FileInfo     : ^aFileInfo;
    BlockInfo    : ^aBlockInfo;
    MaxUsedBlock : Word;

{***************************************************************************}

PROCEDURE XMemShowSize;
{$IFDEF German}
CONST Txt1 = 'benutzt';
{$ELSE}
CONST Txt1 = 'used';
{$ENDIF}
BEGIN
  If XMemInitialized=xmiYes Then
    WriteLn(' -> '+ST( MaxUsedBlock*BlockSizeK,1) + ' kb '+Txt1);
END;

{***************************************************************************}

PROCEDURE XInit;
VAR i: Word;
BEGIN
  If xMemInitialized=xmiNo Then xMemInitialized:=xmiAlloc;

  GetMem(XBuffer, Blocksize);
  GetMem(Blockinfo, xMemCount*SizeOf(tBlockInfo));
  New(Fileinfo);

  i:=0;
  while i<maxfiles do begin
    fileinfo^[i].used:=false;
    inc(i);
  end;

  i:=0;
  while i<xmemcount do begin
    blockinfo^[i].used:=false;
    blockinfo^[i].next:=unused;
    inc(i);
  end;

  MaxUsedBlock:=0;
END;

{***************************************************************************}

PROCEDURE XClose;
BEGIN
{ evtl. offene FileInfo's flushen ? }
  If xMemInitialized=xmiNo Then Exit;

  Dispose(Fileinfo);
  FreeMem(Blockinfo, XmemCount*SizeOf(tBlockInfo));
  FreeMem(XBuffer, BlockSize);
END;

{***************************************************************************}

FUNCTION GetFileInfo(Var F: File; Var fi: Byte): Boolean;
VAR i: Byte;
BEGIN
  fi:=MaxFiles;
  i:=0;
  While i<MaxFiles Do Begin
    If (FileInfo^[i].Used) And
     (FileRec(FileInfo^[i].F^).Handle=FileRec(F).Handle) Then Begin
      fi:=i;
      i:=MaxFiles;
    End;
    Inc(i);
  End;

  GetFileInfo:=(fi<>MaxFiles);
END;

{***************************************************************************}

PROCEDURE FlushMem(fi: Byte);
VAR b: Word;
    w: Boolean;
BEGIN
  w:=(FileInfo^[fi].BCount>=8);

  If w Then OpenCacheFlushWindow;

  While FileInfo^[fi].BCount>0 Do Begin
    b:=FileInfo^[fi].StartBlock;
    If BlockInfo^[b].Dirty Then Begin

      Seek(FileInfo^[fi].F^, FileInfo^[fi].FStart);
      GetError;

      XMemRead(b, XBuffer^);
      BlockWrite(FileInfo^[fi].F^, XBuffer^, BlockSize);
      GetError;

{$IFNDEF OS2}
      Asm
        mov ax, $4A10
        mov bx, $0001
        int $2f
      End;
{$ENDIF}
    End;

    Inc(FileInfo^[fi].FStart, BlockSize);
    FileInfo^[fi].StartBlock:=BlockInfo^[b].Next;
    BlockInfo^[b].Used:=False;
    BlockInfo^[b].Next:=unused;

    Dec(FileInfo^[fi].BCount);
  End;

  If w Then CloseWindow;
END;

{***************************************************************************}

PROCEDURE FlushSomeMem(fi: Byte);
VAR i: Byte;
BEGIN
  For i:=0 To MaxFiles-1 Do
    If (i<>fi) And (FileInfo^[i].Used) And (FileInfo^[i].Mode=xcWrite)
      And ( (FileInfo^[fi].Mode=xcRandom) Or (FileInfo^[i].BCount>FileInfo^[fi].BCount) )
        Then fi:=i;

  If FileInfo^[fi].Mode=xcWrite Then FlushMem(fi);
END;

{***************************************************************************}

FUNCTION GetBlock(fi: Byte; Var bi: Word): Boolean;
{ sucht freien Block, macht ggf. Block frei}
BEGIN
  bi:=0;
  While (bi<XMemCount) And (BlockInfo^[bi].Used) Do
    Inc(bi);

  If bi>=XMemCount Then Begin
    FlushSomeMem(fi);

    bi:=0;
    While (bi<XMemCount) And (BlockInfo^[bi].Used) Do
      Inc(bi);

    If bi>=XMemCount Then GetBlock:=False;
  End Else Begin
    If bi+1>MaxUsedBlock Then MaxUsedBlock:=bi+1;
    GetBlock:=True;
  End;
END;

{***************************************************************************}
{***************************************************************************}

PROCEDURE xMemOpenFile(Var F: File; Mode: Byte; Var Buffer; Var BPos: Word);
{ initalisiert Cache fr File "F". Im Modus xcWrite wird "Buffer" mit
  Clustergrenze beschrieben }

  PROCEDURE stdOpen;
  VAR l: LongInt;
  BEGIN
    l:=Addr2BlockAddr(FileSize(F));
    BPos:=FileSize(F)-l;

    If BPos>0 Then Begin
      Seek(F, l);
      BlockRead(F, Buffer, BPos);
      GetError;
    End;
  END;

VAR i: Byte;
BEGIN
  Result:=0;

  i:=0;
  While (i<MaxFiles) And (FileInfo^[i].Used) Do
    Inc(i);

  If i=MaxFiles Then ErrorAbort:=eaInternalError
  Else Begin
    FileInfo^[i].Used:=True;
    FileInfo^[i].F:=@F;
    FileInfo^[i].Mode:=Mode;
    FileInfo^[i].StartBlock:=unused;
    FileInfo^[i].BCount:=0;

    If xMemInitialized=xmiYes Then Begin
      If Mode=xcWrite Then Begin
        FileInfo^[i].FEnd:=FileSize(F);
        FileInfo^[i].FStart:=Addr2BlockAddr(FileInfo^[i].FEnd);
        BPos:=FileInfo^[i].FEnd-FileInfo^[i].FStart;
        FileInfo^[i].FEnd:=FileInfo^[i].FStart;

        If BPos>0 Then Begin
          Seek(F, FileInfo^[i].FStart);
          BlockRead(F, Buffer, BPos);
          GetError;
        End;
      End Else Begin
        FileInfo^[i].FStart:=0;
        FileInfo^[i].FEnd:=0;
      End;
    End Else If Mode=xcWrite Then stdOpen;
  End;
END;

{***************************************************************************}

PROCEDURE AddBlock(fi: Byte; bi: Word);
{ Dirty und Filled mssen noch gesetzt werden }
VAR bj : Word;
BEGIN
  BlockInfo^[bi].Used:=True;
  BlockInfo^[bi].Next:=unused;

  Inc(FileInfo^[fi].BCount);
  If FileInfo^[fi].BCount=1 Then FileInfo^[fi].StartBlock:=bi
  Else Begin
    bj:=FileInfo^[fi].StartBlock;
    While (BlockInfo^[bj].Next<>unused) And (ErrorAbort=eaOK) Do Begin
      bj:=BlockInfo^[bj].Next;
      If bi=bj Then ErrorAbort:=eaInternalError;
    End;

    BlockInfo^[bj].Next:=bi;
  End;

  If FileInfo^[fi].BCount=1 Then FileInfo^[fi].FStart:=FileInfo^[fi].FEnd;
  Inc(FileInfo^[fi].FEnd, BlockSize);
END;

{***************************************************************************}

PROCEDURE FindBlock(fi: Byte; FPos: LongInt; Var bi: Word);
VAR l: LongInt;
BEGIN
  bi:=FileInfo^[fi].StartBlock;
  l:=FileInfo^[fi].FStart;

  While l<FPos Do Begin
    Inc(l, BlockSize);
    bi:=BlockInfo^[bi].Next;
  End;
END;

{***************************************************************************}

PROCEDURE xMemAppendFile(Var F: File; Var Buffer);
VAR fi : Byte;
    bi : Word;
BEGIN
  Result:=0;
  If (xMemInitialized=xmiYes) And GetFileInfo(F, fi) Then Begin
    If GetBlock(fi, bi) Then Begin
      AddBlock(fi, bi);
      BlockInfo^[bi].Dirty:=True;
      BlockInfo^[bi].Filled:=True;

      XMemWrite(bi, Buffer);
    End Else Begin
      Seek(F, FileInfo^[fi].FEnd);
      GetError;

      BlockWrite(F, Buffer, BlockSize);
      GetError;

      Inc(FileInfo^[fi].FEnd, BlockSize); { Zeiger auf Dateiende }
    End;
  End Else Begin
    BlockWrite(F, Buffer, BlockSize);
    GetError;
  End;
END;

{***************************************************************************}

PROCEDURE xMemTruncFile(Var F: File; FPos: LongInt; Var Buffer; Var BPos: Word);
VAR l      : LongInt;
    bi, bj : Word;
    fi, bc : Byte;
BEGIN
  Result:=0;
  If (xMemInitialized=xmiYes) And GetFileInfo(F, fi) Then Begin
    If FPos>=FileInfo^[fi].FEnd Then BPos:=FPos-FileInfo^[fi].FEnd
    Else Begin
      If FPos>=FileInfo^[fi].FStart Then Begin
        l:=Addr2BlockAddr(FPos);
        BPos:=FPos-l;

        FindBlock(fi, l, bi);

        If BlockInfo^[bi].Filled Then XMemRead(bi, Buffer)
        Else Begin
          Seek(F, l);
          GetError;
          BlockRead(F, Buffer, BPos);
          GetError;
        End;

        { Restspeicher wegwerfen }
        FileInfo^[fi].FEnd:=l;
        Repeat
          bj:=BlockInfo^[bi].Next;
          BlockInfo^[bi].Used:=False;
          BlockInfo^[bi].Next:=unused;
          Dec(FileInfo^[fi].BCount);
          bi:=bj;
        Until bi=Unused;

        If FileInfo^[fi].BCount>0 Then Begin
          FindBlock(fi, l-BlockSize, bi);
          BlockInfo^[bi].Next:=unused;
        End;
      End Else Begin
        { Speicher wegwerfen }
        bi:=FileInfo^[fi].StartBlock;
        While FileInfo^[fi].BCount>0 Do Begin
          bj:=BlockInfo^[bi].Next;
          BlockInfo^[bi].Used:=False;
          BlockInfo^[bi].Next:=unused;
          Dec(FileInfo^[fi].BCount);
          bi:=bj;
        End;

        l:=Addr2BlockAddr(FPos);
        BPos:=FPos-l;
        Seek(F, l);
        GetError;

        BlockRead(F, Buffer, BPos);
        GetError;

        FileInfo^[fi].FStart:=l;
        FileInfo^[fi].FEnd:=l;
      End;

      Seek(F, l);
      GetError;
      Truncate(F);
      GetError;
    End;
  End Else Begin { kein XMem-Speicher }
    If FPos>=FilePos(F) Then BPos:=FPos-FilePos(F)
    Else Begin
      l:=Addr2BlockAddr(FPos);
      BPos:=FPos-l;
      Seek(F, l);
      GetError;

      BlockRead(F, Buffer, BPos);
      GetError;

      Seek(F, l);
      GetError;
      Truncate(F);
      GetError;
    End;
  End;
END;

{***************************************************************************}

PROCEDURE xMemReplaceFile(Var F: File; von, bis: LongInt;
                          Var OrgBuf; OrgSize: Word;
                          Var Buffer; Var BPos: Word);
{ lscht Bereich zwischen "von" und "bis" }
{ DateiEnde liegt bei FileEnd+BPos }
VAR XPos, i    : Word;
    BBuffer    : TField absolute Buffer;
    ResultSize : LongInt;
    CutSize    : LongInt;
    l, n, v    : LongInt;
BEGIN
  Result:=0;

  CutSize:=bis-von;
  ResultSize:=xMemFilePos(F) + BPos - CutSize + OrgSize;
  l:=xMemFilePos(F) + BPos - bis; { soviel }

  xMemAppendFile(F, Buffer);

  If CutSize>OrgSize Then Begin
    n:=von+OrgSize; { hierhin }
    v:=bis;         { hierher }

    While (l>0) Do Begin
      xMemReadFile(F, Addr2BlockAddr(n), BlockSize, XBuffer^);
      xMemReadFile(F, Addr2BlockAddr(v), BlockSize, BBuffer);

      XPos:=n-Addr2BlockAddr(n);
      BPos:=v-Addr2BlockAddr(v);

      If XPos>BPos Then i:=BlockSize-XPos
      Else i:=BlockSize-BPos;

      Move(BBuffer[BPos], XBuffer^[XPos], i);

      xMemWriteFile(F, Addr2BlockAddr(n), BlockSize, XBuffer^);

      Dec(l, i);
      Inc(n, i);
      Inc(v, i);
    End;
  End Else If CutSize<OrgSize Then Begin
    v:=xMemFilePos(F)-(BlockSize-BPos); { hierher }
    n:=v + (OrgSize-CutSize);           { hierhin }

    While (l>0) Do Begin
      xMemReadFile(F, Addr2BlockAddr(n), BlockSize, XBuffer^);
      xMemReadFile(F, Addr2BlockAddr(v), BlockSize, BBuffer);

      XPos:=n-Addr2BlockAddr(n);
      BPos:=v-Addr2BlockAddr(v);

      If XPos>BPos Then i:=BlockSize-XPos
      Else i:=BlockSize-BPos;

      Move(BBuffer[BPos], XBuffer^[XPos], i);

      xMemWriteFile(F, Addr2BlockAddr(n), BlockSize, XBuffer^);

      Dec(l, i);
      Dec(n, i);
      Dec(v, i);
    End;
  End;

  { OrgBuf einfgen }
  l:=Addr2BlockAddr(von);
  XPos:=von-l; { Zeichen in XBuffer, die bleiben sollen }
  xMemReadFile(F, l, BlockSize, XBuffer^);
  { kann Dateilesefehler ber Dateiende verursachen ! }

  i:=BlockSize-XPos;
  If i>=OrgSize Then Move(OrgBuf, XBuffer^[XPos], OrgSize)
  Else Begin
    Move(OrgBuf, XBuffer^[XPos], i);
    xMemWriteFile(F, l, BlockSize, XBuffer^);

    Inc(l, BlockSize);
    xMemReadFile(F, l, BlockSize, XBuffer^);
    Move(tField(OrgBuf)[i], XBuffer^, OrgSize-i);
  End;
  xMemWriteFile(F, l, BlockSize, XBuffer^);

  xMemTruncFile(F, ResultSize, Buffer, BPos);
END;

{***************************************************************************}

PROCEDURE xMemReadFile(Var F: File; Pos: LongInt; Size: Word; Var Buffer);
{ Pos mu an Clustergrenze liegen !! }
VAR l      : LongInt;
    bi     : Word;
    fi     : Byte;
BEGIN
  Result:=0;
  If (XMemInitialized=xmiYes) And GetFileInfo(F, fi) Then Begin
    If (Pos>=FileInfo^[fi].FStart) And (Pos<FileInfo^[fi].FEnd) Then Begin
      { aus dem Speicher lesen }
      FindBlock(fi, Pos, bi);

      If BlockInfo^[bi].Filled Then XMemRead(bi, Buffer)
      Else Begin
        Seek(F, Pos);
        GetError;
        BlockRead(F, Buffer, Size);
        GetError;

        XMemWrite(bi, Buffer);
        BlockInfo^[bi].Filled:=True;
      End;
    End Else Begin
      { von Disk lesen }
      Seek(F, Pos);
      GetError;
      BlockRead(F, Buffer, Size);
      GetError;

      { in Speicher kopieren ? }
      If FileInfo^[fi].Mode=xcRandom Then
        While (Pos>=FileInfo^[fi].FEnd) And (GetBlock(fi, bi)) Do Begin
          AddBlock(fi, bi);

          If Pos<FileInfo^[fi].FEnd Then Begin
            XMemWrite(bi, Buffer);
            BlockInfo^[bi].Dirty:=True;
            BlockInfo^[bi].Filled:=True;
          End Else Begin
            BlockInfo^[bi].Dirty:=False;
            BlockInfo^[bi].Filled:=False;
          End;
        End;
    End;
  End Else Begin
    Seek(F, Pos);
    GetError;
    BlockRead(F, Buffer, Size);
    GetError;
  End;
END;

{***************************************************************************}

PROCEDURE xMemWriteFile(Var F: File; Pos: LongInt; Size: Word; Var Buffer);
{ Pos mu an Clustergrenze liegen !! }
VAR l      : LongInt;
    bi, bj : Word;
    fi, i  : Byte;
BEGIN
  Result:=0;
  If (XMemInitialized=xmiYes) And GetFileInfo(F, fi) Then Begin
    If (Pos>=FileInfo^[fi].FStart) And (Pos<FileInfo^[fi].FEnd) Then Begin
      { Speicher aktualisieren }
      FindBlock(fi, Pos, bi);
      XMemWrite(bi, Buffer);
      BlockInfo^[bi].Dirty:=True;
      BlockInfo^[bi].Filled:=True;
    End Else Begin
      { Speicher anfordern }
      If FileInfo^[fi].Mode=xcRandom Then
        While (Pos>=FileInfo^[fi].FEnd) And (GetBlock(fi, bi)) Do Begin
          AddBlock(fi, bi);

          { FEnd wurde vergrert }
          If Pos<FileInfo^[fi].FEnd Then Begin
            XMemWrite(bi, Buffer);
            BlockInfo^[bi].Dirty:=True;
            BlockInfo^[bi].Filled:=True;
          End Else Begin
            BlockInfo^[bi].Dirty:=False;
            BlockInfo^[bi].Filled:=False;
          End;
        End;

      If FileInfo^[fi].FEnd<=Pos Then Begin
        Seek(F, Pos);
        GetError;
        BlockWrite(F, Buffer, Size);
        GetError;
      End;
    End;
  End Else Begin
    Seek(F, Pos);
    GetError;
    BlockWrite(F, Buffer, Size);
    GetError;
  End;
END;

{***************************************************************************}

FUNCTION xMemFilePos(Var F: File): LongInt;
VAR l  : LongInt;
    fi : Byte;
BEGIN
  If (XMemInitialized=xmiYes) And GetFileInfo(F, fi) Then Begin
    xMemFilePos:=FileInfo^[fi].FEnd;
  End Else Begin
    l:=FilePos(F);
    GetError;
    xMemFilePos:=l;
  End;
END;

{***************************************************************************}

PROCEDURE xMemCloseFile(Var F: File; Var Buffer; Size: Word);
VAR fi: Byte;
BEGIN
  Result:=0;
  If (XMemInitialized=xmiYes) And GetFileInfo(F, fi) Then Begin
    FlushMem(fi);
    Seek(F, FileInfo^[fi].FEnd);
    GetError;
  End;

  If Size>0 Then Begin
    BlockWrite(F, Buffer, Size);
    GetError;
  End;

  GetFileInfo(F, fi);
  If FileInfo^[fi].Mode=xcWrite Then Begin
    Truncate(F);
    GetError;
  End;

  FileInfo^[fi].Used:=False;
END;

{***************************************************************************}

END.
